import errno
import pickle

from flask import current_app, render_template
from flask.views import MethodView
from werkzeug.exceptions import Forbidden

from ..constants import ID, TIMESTAMP_UPLOAD
from ..utils.date_funcs import delete_if_lifetime_over
from ..utils.permissions import LIST, may


def file_infos(names=None):
    """
    iterates over storage files metadata.
    note: we put the storage name into the metadata as ID

    :param names: None means "all items"
                  otherwise give a list of storage item names
    """
    storage = current_app.storage
    if names is None:
        names = list(storage)
    for name in names:
        try:
            with storage.open(name) as item:
                meta = dict(item.meta)
                if not meta:
                    # we got empty metadata, this happens for 0-byte .meta files.
                    # ignore it for now.
                    continue
                if delete_if_lifetime_over(item, name):
                    continue
                meta[ID] = name
                yield meta
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        except pickle.UnpicklingError:
            # corrupted meta file, just ignore it for now
            pass


class FileListView(MethodView):
    def get(self):
        if not may(LIST):
            raise Forbidden()
        files = sorted(file_infos(), key=lambda f: f[TIMESTAMP_UPLOAD], reverse=True)
        return render_template('filelist.html', files=files)



#
# REST api tests
#

import os
import base64
import threading
import time
import copy
import hashlib
import re
from requests.auth import _basic_auth_str
from flask import current_app, url_for, json

import pytest

from ..app import create_app
from ..config import Config
from ..constants import FILENAME, TYPE, LOCKED, SIZE, COMPLETE, HASH, \
    TIMESTAMP_DOWNLOAD, TIMESTAMP_UPLOAD, TIMESTAMP_MAX_LIFE, TRANSACTION_ID
from ..utils.date_funcs import get_maxlife

UPLOAD_DATA = b"""\
#!/usr/bin/python3

print('hello, world')
"""


class FakeTime:
    """Overwrite time.time to control the timestamp that is used in server"""

    def __init__(self, now=None):
        self.orig = time.time
        self.now = now

    def set_time(self, now=None):
        self.now = now
        if now and time.time != self.get_time:
            time.time = self.get_time
        elif now is None and time.time == self.get_time:
            time.time = self.orig

    def get_time(self):
        return self.now

    def __enter__(self):
        self.set_time(self.now)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.set_time(None)


def wait_background():
    """Wait until background threads terminate."""

    main_thread = threading.current_thread()
    for t in threading.enumerate():
        if t is not main_thread:
            t.join()


@pytest.fixture
def client_fixture(tmp_path):
    with FakeTime() as faketime:
        Config.PERMISSIONS = {
            'admin': 'admin,list,create,modify,read,delete',
            'full': 'list,create,modify,read,delete',
            'none': '',
        }
        Config.STORAGE_FILESYSTEM_DIRECTORY = str(tmp_path)

        app = create_app()
        app.config['TESTING'] = True

        with app.test_client() as client:
            with app.app_context():
                yield app, client, faketime

        wait_background()


class TmpConfig:
    def __init__(self, app, tmp_config=None):
        self.config = app.config
        self.orig_config = copy.deepcopy(app.config)
        self.tmp_config = {} if tmp_config is None else tmp_config

    def __enter__(self):
        self.config.update(self.tmp_config)

    def __exit__(self, exc_type, exc_value, traceback):
        self.config.update(self.orig_config)


class RestUrl:
    def __init__(self, item_id=None):
        self.item_id = item_id

    @property
    def config(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.api_info')

    @property
    def upload(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items')

    @property
    def list(self):
        return self.upload

    @property
    def detail(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_detail', name=self.item_id)

    @property
    def download(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_download', name=self.item_id)

    @property
    def delete(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_delete', name=self.item_id)

    @property
    def modify(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_modify', name=self.item_id)

    @property
    def lock(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_lock', name=self.item_id)

    @property
    def unlock(self):
        with current_app.test_request_context():
            return url_for('bepasty_apis.items_unlock', name=self.item_id)


def add_auth(user, password, headers=None):
    headers = headers if headers is not None else {}
    headers['Authorization'] = _basic_auth_str(user, password)
    return headers


def test_invalid_url(client_fixture):
    _, client, _ = client_fixture

    response = client.get('/apis/rest/invalid')
    assert response.status_code == 404


def check_response(response, code, ftype='application/json', check_data=True):
    assert response.status_code == code
    assert response.headers['Content-Type'] == ftype
    if check_data:
        assert int(response.headers['Content-Length']) == len(response.data)


def check_err_response(response, code, check_data=True):
    check_response(response, code, check_data=check_data)
    if check_data:
        assert code == response.json['error']['code']
    # check if doesn't have html tag
    assert not re.match(r'<.+>', response.data.decode())


def check_data_response(response, meta, data, offset=0, total_size=None,
                        check_data=True):
    ftype = meta['file-meta'][TYPE]
    filename = meta['file-meta'][FILENAME]
    if total_size is None:
        total_size = len(data)

    disposition = 'attachment; filename="{}"'.format(filename)
    range_str = 'bytes {}-{}/{}'.format(offset, offset + len(data) - 1,
                                        total_size)

    check_response(response, 200, ftype, check_data)
    assert response.headers['Content-Disposition'] == disposition
    assert response.headers['Content-Range'] == range_str
    if check_data:
        assert data == response.data


def check_json_response(response, metas, code=200, check_data=True):
    check_response(response, code, check_data=check_data)
    if check_data:
        assert metas == response.json


def check_upload_response(response, code=201, check_data=True):
    check_json_response(response, {}, code=code, check_data=check_data)

    if code == 200:
        assert len(response.headers[TRANSACTION_ID]) > 0
        return None

    url_prefix = RestUrl().upload + '/'
    assert response.headers['Content-Location'].startswith(url_prefix)
    return response.headers['Content-Location']


def test_auth(client_fixture):
    _, client, _ = client_fixture
    url = RestUrl()

    # basic auth (unknown user)
    response = client.get(url.list, headers=add_auth('user', 'invalid'))
    check_err_response(response, 403)
    # basic auth (valid user)
    response = client.get(url.list, headers=add_auth('user', 'full'))
    check_response(response, 200)

    # token auth (unknown user)
    response = client.get(url.list + '?token=invalid')
    check_err_response(response, 403)
    # token auth (valid user)
    response = client.get(url.list + '?token=full')
    check_response(response, 200)


def test_config(client_fixture):
    app, client, _ = client_fixture

    url = RestUrl()

    # get server config (post should fail)
    response = client.post(url.config)
    check_err_response(response, 405)

    # get server config
    response = client.get(url.config)
    check_response(response, 200)
    assert len(response.json) == 2
    assert response.json['MAX_ALLOWED_FILE_SIZE'] == app.config['MAX_ALLOWED_FILE_SIZE']
    assert response.json['MAX_BODY_SIZE'] == app.config['MAX_BODY_SIZE']

    # get server config (head)
    response = client.head(url.config)
    check_response(response, 200, check_data=False)


def _upload(client, data, token=None, filename=None, ftype=None, lifetime=None,
            range_str=None, trans_id=None, encode=True, set_range=True):
    if data:
        payload = base64.b64encode(data) if encode else data
        payload_len = len(payload)
    else:
        payload = None
        payload_len = 0

    headers = {
        'Content-Length': str(payload_len),
    }
    if set_range:
        if range_str is None:
            range_str = 'bytes 0-{}/{}'.format(payload_len - 1, payload_len)
        headers['Content-Range'] = range_str
    if filename is not None:
        headers['Content-Filename'] = filename
    if ftype is not None:
        headers['Content-Type'] = ftype
    if lifetime:
        headers['Maxlife-Value'] = str(lifetime[0])
        headers['Maxlife-Unit'] = lifetime[1]
    if trans_id is not None:
        headers[TRANSACTION_ID] = trans_id
    if token is not None:
        add_auth('user', token, headers)

    response = client.post(RestUrl().upload, headers=headers, data=payload)
    # FIXME: without waiting for background hash compute, following
    # detail request may not have hash yet
    wait_background()

    return response


def make_meta(data, filename=None, ftype=None, lifetime=None, uri=None):
    h = hashlib.sha256()
    h.update(data)

    if lifetime is None:
        # default maxlife is 1 MONTHS
        lifetime = [1, 'MONTHS']

    maxtime = get_maxlife({
        'maxlife_value': lifetime[0],
        'maxlife_unit': lifetime[1]
    }, True)
    maxlife = int(time.time()) + maxtime if maxtime > 0 else maxtime

    meta = {
        'file-meta': {
            COMPLETE: True,
            FILENAME: filename,
            HASH: h.hexdigest(),
            LOCKED: False,
            SIZE: len(data),
            TIMESTAMP_DOWNLOAD: 0,
            TIMESTAMP_MAX_LIFE: maxlife,
            TIMESTAMP_UPLOAD: int(time.time()),
            TYPE: ftype,
        },
        'uri': uri,
    }
    return meta


def upload(client, data, token=None, filename=None, ftype=None, lifetime=None):
    response = _upload(client, data, token, filename, ftype, lifetime)
    uri = check_upload_response(response)

    return make_meta(data, filename, ftype, lifetime, uri)


def upload_files(client):
    datas = {}
    metas = {}

    for i in (1, 2):

        data = """\
#!/usr/bin/python3

print('hello,world {}')
""".format(i).encode()

        filename = '{}-test.py'.format(i)
        lifetime = [1, 'YEARS']
        ftype = 'text/x-python'

        with FakeTime(int(time.time()) + i):
            meta = upload(client, data, token='full', filename=filename,
                          ftype=ftype, lifetime=lifetime)

        item_id = os.path.basename(meta['uri'])
        datas[item_id] = data
        metas[item_id] = meta

    assert len(datas) == 2
    assert len(metas) == 2

    return datas, metas


def check_detail_or_download(app, client, item_id, meta, data, download=False):
    url = RestUrl(item_id=item_id)

    url = url.download if download else url.detail

    # post should fail
    response = client.post(url)
    check_err_response(response, 405)

    # get with default (no permission)
    response = client.get(url)
    check_err_response(response, 403)
    # head with default (no permission)
    response = client.head(url)
    check_err_response(response, 403, check_data=False)

    with TmpConfig(app, {'DEFAULT_PERMISSIONS': 'read'}):
        # head with default (has permission)
        response = client.head(url)
        if download:
            check_data_response(response, meta, data, check_data=False)
        else:
            check_json_response(response, meta, check_data=False)

        # get with default (has permission)
        response = client.get(url)
        if download:
            check_data_response(response, meta, data)
        else:
            check_json_response(response, meta)

    # get with no permission
    response = client.get(url, headers=add_auth('user', 'none'))
    check_err_response(response, 403)
    # head with no permission
    response = client.head(url, headers=add_auth('user', 'none'))
    check_err_response(response, 403, check_data=False)

    # head with permission
    response = client.head(url, headers=add_auth('user', 'full'))
    if download:
        check_data_response(response, meta, data, check_data=False)
    else:
        check_json_response(response, meta, check_data=False)

    # get with permission
    response = client.get(url, headers=add_auth('user', 'full'))
    if download:
        check_data_response(response, meta, data)
    else:
        check_json_response(response, meta)


def check_expired(client, item_id, download=False):
    url = RestUrl(item_id=item_id)
    url = url.download if download else url.detail

    # expired items
    response = client.get(url, headers=add_auth('user', 'full'))
    check_err_response(response, 404)


def test_upload_basic(client_fixture):
    _, client, _ = client_fixture

    # simple upload without permission
    response = _upload(client, UPLOAD_DATA, token='none', filename='test.py',
                       ftype='text/x-python', lifetime=[1, 'FOREVER'])
    check_err_response(response, 403)

    # simple upload with permission
    upload(client, UPLOAD_DATA, token='full', filename='test.py',
           ftype='text/x-python', lifetime=[1, 'FOREVER'])


def test_upload_params(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    for filename, ftype in ((None, None), ('test.py', None), (None, 'text/plain')):
        # upload without filename and/or ftype
        meta = upload(client, UPLOAD_DATA, token='full',
                      filename=filename, ftype=ftype)
        item_id = os.path.basename(meta['uri'])

        url = RestUrl(item_id)
        # get meta for uploaded item
        response = client.get(url.detail, headers=add_auth('user', 'full'))
        check_json_response(response, None, check_data=False)
        assert len(response.json) > 0
        # copy generated meta
        if filename:
            assert response.json['file-meta'][FILENAME] == filename
        else:
            assert len(response.json['file-meta'][FILENAME]) > 0
            meta['file-meta'][FILENAME] = response.json['file-meta'][FILENAME]
        if ftype:
            assert response.json['file-meta'][TYPE] == ftype
        else:
            assert len(response.json['file-meta'][TYPE]) > 0
            meta['file-meta'][TYPE] = response.json['file-meta'][TYPE]

        check_detail_or_download(app, client, item_id, meta, UPLOAD_DATA)
        check_detail_or_download(app, client, item_id, meta, UPLOAD_DATA,
                                 download=True)


def test_upload_range(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    filename = 'test.py'
    ftype = 'text/x-python'

    sep = 10
    data = UPLOAD_DATA[:sep]

    # (invalid no Content-Range:)
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, set_range=False)
    check_err_response(response, 400)

    # Content-Range: invalid (invalid format)
    range_str = 'invalid'
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes invalid (invalid format)
    range_str = 'bytes invalid'
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: other 0-<sep - 1>/<len(data)> (invalid unit)
    range_str = 'other {}-{}/{}'.format(0, sep - 1, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes invalid-<sep - 1>/<len(data)> (invalid first)
    range_str = 'bytes invalid-{}/{}'.format(sep - 1, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes 0-invalid/<len(data)> (invalid last)
    range_str = 'bytes {}-invalid/{}'.format(0, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes <sep - 1>-0/<len(data)> (invalid first > last)
    range_str = 'bytes {}-{}/{}'.format(sep - 1, 0, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes 0-<sep - 1>/* (not supported)
    range_str = 'bytes {}-{}/*'.format(0, sep - 1)
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes */<len(data)> (not supported)
    range_str = 'bytes */{}'.format(len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # Content-Range: bytes */* (not supported)
    range_str = 'bytes */*'
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_err_response(response, 400)

    # upload first part of data
    range_str = 'bytes {}-{}/{}'.format(0, sep - 1, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_upload_response(response, 200)

    # upload remainder of data
    data = UPLOAD_DATA[sep:]
    range_str = 'bytes {}-{}/{}'.format(sep, len(UPLOAD_DATA) - 1,
                                        len(UPLOAD_DATA))
    response = _upload(client, data, token='full', range_str=range_str,
                       trans_id=response.headers[TRANSACTION_ID])
    uri = check_upload_response(response)
    item_id = os.path.basename(uri)

    # check a uploaded item
    meta = make_meta(UPLOAD_DATA, filename=filename, ftype=ftype, uri=uri)
    check_detail_or_download(app, client, item_id, meta, UPLOAD_DATA)
    check_detail_or_download(app, client, item_id, meta, UPLOAD_DATA,
                             download=True)

    # upload again with item.data.size != start position (invalid position)
    data = UPLOAD_DATA[sep + 1:]
    range_str = 'bytes {}-{}/{}'.format(sep + 1, len(UPLOAD_DATA) - 1,
                                        len(UPLOAD_DATA))
    response = _upload(client, data, token='full', range_str=range_str,
                       trans_id=response.headers[TRANSACTION_ID])
    check_err_response(response, 409)


def test_bad_data(client_fixture):
    app, client, _ = client_fixture

    filename = 'test.py'
    ftype = 'text/x-python'

    # upload without Content-Length
    range_str = 'bytes 0-{}/{}'.format(len(UPLOAD_DATA) - 1, len(UPLOAD_DATA))
    response = _upload(client, None, token='full', filename=filename,
                       ftype=ftype, range_str=range_str, encode=False)
    check_err_response(response, 400)

    # upload invalid base64 encode
    range_str = 'bytes 0-{}/{}'.format(len(UPLOAD_DATA) - 1, len(UPLOAD_DATA))
    response = _upload(client, UPLOAD_DATA, token='full', filename=filename,
                       ftype=ftype, range_str=range_str, encode=False)
    check_err_response(response, 400)

    # server must not have left garbage files
    assert len(os.listdir(app.config['STORAGE_FILESYSTEM_DIRECTORY'])) == 0


def test_upload_maxlife(client_fixture):
    """upload maxlife"""

    _, client, _ = client_fixture

    lifetime = ['foo', 'MINUTES']
    response = _upload(client, UPLOAD_DATA, token='full', lifetime=lifetime)
    assert response.status_code == 400

    lifetime = ['1', 'foo']
    response = _upload(client, UPLOAD_DATA, token='full', lifetime=lifetime)
    assert response.status_code == 400


def test_upload_transaction_id(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    sep = 10
    data = UPLOAD_DATA[sep:]
    range_str = 'bytes {}-{}/{}'.format(sep, len(UPLOAD_DATA) - 1,
                                        len(UPLOAD_DATA))

    # upload with invalid Transaction-ID
    trans_id = 'invalid'
    response = _upload(client, data, token='full', range_str=range_str,
                       trans_id=trans_id)
    check_err_response(response, 400)

    # upload with Transaction-ID for invalid filename
    trans_id = base64.b64encode(b'invalid')
    response = _upload(client, data, token='full', range_str=range_str,
                       trans_id=trans_id)
    check_err_response(response, 400)


def test_upload_too_big(client_fixture):
    app, client, _ = client_fixture

    # upload too big size
    size = 10000
    with TmpConfig(app, {'MAX_ALLOWED_FILE_SIZE': size}):
        ftype = 'text/x-bepasty-list'
        data = bytes(b'a' * (size + 10))
        response = _upload(client, data, token='full', ftype=ftype)
        check_err_response(response, 413)


def test_list_basic(client_fixture):
    app, client, faketime = client_fixture

    url = RestUrl()
    faketime.set_time(100)

    datas, metas = upload_files(client)

    # list with default (no permission)
    response = client.get(url.list)
    check_err_response(response, 403)
    response = client.head(url.list)
    check_err_response(response, 403, check_data=False)

    # list with default (has permission)
    with TmpConfig(app, {'DEFAULT_PERMISSIONS': 'list'}):
        response = client.get(url.list)
        check_json_response(response, metas)
        response = client.head(url.list)
        check_json_response(response, metas, check_data=False)

    # list with no permission
    response = client.get(url.list, headers=add_auth('user', 'none'))
    check_err_response(response, 403)
    response = client.head(url.list, headers=add_auth('user', 'none'))
    check_err_response(response, 403, check_data=False)

    # list with permission
    response = client.get(url.list, headers=add_auth('user', 'full'))
    check_json_response(response, metas)
    response = client.head(url.list, headers=add_auth('user', 'full'))
    check_json_response(response, metas, check_data=False)

    # adjust time to exceed lifetime
    faketime.set_time(3600 * 24 * 365 * 2)
    # list for expired items
    response = client.get(url.list, headers=add_auth('user', 'full'))
    check_json_response(response, {})


def test_detail_basic(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    datas, metas = upload_files(client)

    for item_id in metas.keys():
        data = datas[item_id]
        meta = metas[item_id]

        check_detail_or_download(app, client, item_id, meta, data)

        # detail of expired item
        faketime.set_time(3600 * 24 * 365 * 2)
        check_expired(client, item_id)
        # check again (should be deleted already)
        faketime.set_time(100)
        check_expired(client, item_id)


def test_download_basic(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    datas, metas = upload_files(client)

    for item_id in metas.keys():
        data = datas[item_id]
        meta = metas[item_id]

        check_detail_or_download(app, client, item_id, meta, data,
                                 download=True)

        # download of expired item
        faketime.set_time(3600 * 24 * 365 * 2)
        check_expired(client, item_id, download=True)
        # check again (should be deleted already)
        faketime.set_time(100)
        check_expired(client, item_id)


def test_download_range(client_fixture):
    _, client, faketime = client_fixture

    faketime.set_time(100)

    datas, metas = upload_files(client)

    for item_id in metas.keys():
        data = datas[item_id]
        meta = metas[item_id]

        url = RestUrl(item_id=item_id)
        headers = add_auth('user', 'full')

        # Range: other (invalid format)
        offset = 0
        limit = 10
        headers['Range'] = 'other'
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=invalid (invalid format)
        offset = 0
        limit = 10
        headers['Range'] = 'bytes=invalid'
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: other=0-10 (invalid unit)
        offset = 0
        limit = 10
        headers['Range'] = 'other={}-{}'.format(offset, limit - 1)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=invalid-10 (invalid first)
        offset = 0
        limit = 10
        headers['Range'] = 'bytes=invalid-{}'.format(limit - 1)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=0-invalid (invalid last)
        offset = 0
        limit = 10
        headers['Range'] = 'other={}-invalid'.format(offset)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=10-0 (invalid first > last)
        offset = 0
        limit = 10
        headers['Range'] = 'bytes={}-{}'.format(limit - 1, offset)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=0-9,10-<limit - 1> (not supported for now)
        offset = 0
        limit = len(data)
        headers['Range'] = 'bytes={}-9,10-{}'.format(offset, limit - 1)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=0-10
        offset = 0
        limit = 10
        headers['Range'] = 'bytes={}-{}'.format(offset, limit - 1)
        response = client.get(url.download, headers=headers)
        check_data_response(response, meta, data[offset:limit], offset=offset,
                            total_size=len(data))

        # Range: bytes=-9
        # FIXME: suffix-byte-range-spec is not supported for now
        offset = 0
        limit = 10
        headers['Range'] = 'bytes=-{}'.format(limit - 1)
        response = client.get(url.download, headers=headers)
        check_err_response(response, 400)

        # Range: bytes=10-<limit - 1>
        offset = 10
        limit = len(data)
        headers['Range'] = 'bytes={}-{}'.format(offset, limit - 1)
        response = client.get(url.download, headers=headers)
        check_data_response(response, meta, data[offset:limit], offset=offset,
                            total_size=len(data))

        # Range: bytes=10-
        offset = 10
        limit = len(data)
        headers['Range'] = 'bytes={}-'.format(offset)
        response = client.get(url.download, headers=headers)
        check_data_response(response, meta, data[offset:limit], offset=offset,
                            total_size=len(data))


def test_modify(client_fixture):
    app, client, _ = client_fixture

    meta = upload(client, UPLOAD_DATA, token='full', filename='test.py',
                  ftype='text/x-python', lifetime=[1, 'FOREVER'])

    item_id = os.path.basename(meta['uri'])
    url = RestUrl(item_id)

    check_detail_or_download(app, client, item_id, meta, None)

    headers = {'Content-Type': 'application/json'}

    # no permission
    response = client.post(url.modify, headers=headers, data='{}')
    check_err_response(response, 403)

    headers = add_auth('user', 'full', headers)

    # invalid name
    response = client.post(RestUrl('abcdefgh').modify, headers=headers, data='{}')
    check_err_response(response, 404)

    # invalid Content-Type
    response = client.post(url.modify, headers=add_auth('user', 'full'), data='{}')
    check_err_response(response, 400)

    # invalid json
    response = client.post(url.modify, headers=headers, data='')
    check_err_response(response, 400)

    # change filename
    filename = 'test2.py'
    meta['file-meta'][FILENAME] = filename
    data = json.dumps({FILENAME: filename})
    response = client.post(url.modify, headers=headers, data=data)
    check_json_response(response, {})

    check_detail_or_download(app, client, item_id, meta, None)

    # change type
    content_type = 'text/plain'
    meta['file-meta'][TYPE] = content_type
    data = json.dumps({TYPE: content_type})
    response = client.post(url.modify, headers=headers, data=data)
    check_json_response(response, {})

    check_detail_or_download(app, client, item_id, meta, None)


def test_delete_basic(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    datas, metas = upload_files(client)

    url = RestUrl('abcdefgh')

    # delete ENOENT item
    response = client.post(url.delete, headers=add_auth('user', 'admin'))
    check_err_response(response, 404)

    for item_id in metas.keys():
        url = RestUrl(item_id)

        # no permission
        response = client.post(url.delete, headers=add_auth('user', 'invalid'))
        check_err_response(response, 403)

        # has permission
        response = client.post(url.delete, headers=add_auth('user', 'full'))
        check_json_response(response, {})

        # should already be deleted
        response = client.post(url.delete, headers=add_auth('user', 'full'))
        check_err_response(response, 404)


def test_lock_basic(client_fixture):
    app, client, faketime = client_fixture

    faketime.set_time(100)

    datas, metas = upload_files(client)

    url = RestUrl('abcdefgh')

    for u in (url.lock, url.unlock):
        # lock/unlock ENOENT item
        response = client.post(u, headers=add_auth('user', 'admin'))
        check_err_response(response, 404)

    for item_id in metas.keys():
        url = RestUrl(item_id)

        for u in (url.lock, url.unlock):
            # lock/unlock, no permission (invalid user)
            response = client.post(u, headers=add_auth('user', 'invalid'))
            check_err_response(response, 403)

            # lock/unlock, no permission (not admin)
            response = client.post(u, headers=add_auth('user', 'full'))
            check_err_response(response, 403)

            # lock/unlock, has permission
            response = client.post(u, headers=add_auth('user', 'admin'))
            check_json_response(response, {})

        # lock item
        response = client.post(url.lock, headers=add_auth('user', 'admin'))
        check_json_response(response, {})

        # download locked item (should fail)
        response = client.get(url.download, headers=add_auth('user', 'full'))
        check_err_response(response, 403)

        # download locked item with admin (should succeed)
        response = client.get(url.download, headers=add_auth('user', 'admin'))
        check_data_response(response, metas[item_id], datas[item_id])

        # modify locked item (should fail)
        headers = add_auth('user', 'full', {'Content-Type': 'application/json'})
        response = client.post(url.modify, headers=headers, data='{}')
        check_err_response(response, 403)

        # modify locked item with admin (should succeed)
        headers = add_auth('user', 'admin', {'Content-Type': 'application/json'})
        response = client.post(url.modify, headers=headers, data='{}')
        check_json_response(response, {})

        # delete locked item (should fail)
        response = client.post(url.delete, headers=add_auth('user', 'full'))
        check_err_response(response, 403)

        # delete locked item with admin (should succeed)
        response = client.post(url.delete, headers=add_auth('user', 'admin'))
        check_json_response(response, {})

        # deleted item
        response = client.post(url.delete, headers=add_auth('user', 'admin'))
        check_err_response(response, 404)


def test_incomplete(client_fixture):
    _, client, _ = client_fixture

    filename = 'test.py'
    ftype = 'text/x-python'

    # upload a half of data to make incomplete
    sep = 10
    data = UPLOAD_DATA[:sep]
    range_str = 'bytes {}-{}/{}'.format(0, sep - 1, len(UPLOAD_DATA))
    response = _upload(client, data, token='full', filename=filename,
                       ftype=ftype, range_str=range_str)
    check_upload_response(response, 200)
    assert len(response.headers[TRANSACTION_ID]) > 0

    # get incomplete item from list
    url = RestUrl()
    headers = add_auth('user', 'full')
    response = client.get(url.list, headers=headers)
    check_response(response, 200)
    assert len(response.json) == 1

    item_id = list(response.json.keys())[0]

    url = RestUrl(item_id)

    # detail should error with incomplete
    response = client.get(url.detail, headers=add_auth('user', 'full'))
    check_err_response(response, 409)

    # download should error with incomplete
    response = client.get(url.download, headers=add_auth('user', 'full'))
    check_err_response(response, 409)

    # modify should error with incomplete
    headers = add_auth('user', 'full', {'Content-Type': 'application/json'})
    response = client.post(url.modify, headers=headers, data='{}')
    check_err_response(response, 409)

    # lock should error with incomplete
    response = client.post(url.lock, headers=add_auth('user', 'admin'))
    check_err_response(response, 409)

    # unlock should error with incomplete
    response = client.post(url.unlock, headers=add_auth('user', 'admin'))
    check_err_response(response, 409)

    # delete should error with incomplete
    response = client.post(url.delete, headers=add_auth('user', 'full'))
    check_err_response(response, 409)

    # delete by admin with incomplete should succeed
    response = client.post(url.delete, headers=add_auth('user', 'admin'))
    check_json_response(response, {})


def test_magic(client_fixture):
    app, client, _ = client_fixture

    try:
        import magic
        assert magic is not None  # suppress flake8 warning
    except ImportError:
        pytest.skip("skipping test, no python-magic installed")
    else:
        with TmpConfig(app, {'USE_PYTHON_MAGIC': True}):
            filename = None
            ftype = None

            # upload a half of data to make incomplete with auto mime
            # detection (meta has 'type-hint' internally)
            sep = 10
            data = UPLOAD_DATA[:sep]
            range_str = 'bytes {}-{}/{}'.format(0, sep - 1, len(UPLOAD_DATA))
            response = _upload(client, data, token='full', filename=filename,
                               ftype=ftype, range_str=range_str)
            check_upload_response(response, 200)
            assert len(response.headers[TRANSACTION_ID]) > 0
            trans_id = response.headers[TRANSACTION_ID]

            # get incomplete item from list
            url = RestUrl()
            headers = add_auth('user', 'full')
            response = client.get(url.list, headers=headers)
            check_response(response, 200)
            assert len(response.json) == 1

            fid = list(response.json.keys())[0]
            meta = list(response.json.values())[0]

            # 'type-hint' should be invisible
            assert meta['file-meta'] is not None
            assert meta['file-meta'].get('type-hint') is None

            # 'type' is 'application/octet-stream' for now
            assert meta['file-meta'][TYPE] == 'application/octet-stream'

            # complete upload
            data = UPLOAD_DATA[sep:]
            range_str = 'bytes {}-{}/{}'.format(sep, len(UPLOAD_DATA) - 1,
                                                len(UPLOAD_DATA))
            response = _upload(client, data, token='full', range_str=range_str,
                               trans_id=trans_id)
            check_upload_response(response)

            # get detail to check python-magic auto detection
            url = RestUrl(fid)
            response = client.get(url.detail, headers=add_auth('user', 'full'))
            check_json_response(response, None, check_data=False)

            assert response.json['file-meta'] is not None
            assert response.json['file-meta'][TYPE] == 'text/x-python'



from bepasty.storage.filesystem import Meta


def test(tmpdir):
    p = tmpdir.join("test.meta")

    m = Meta(p.open('w+b'))
    assert len(m) == 0
    m.close()

    m = Meta(p.open('r+b'))
    m['flag'] = True
    assert len(m) == 1
    m.close()

    m = Meta(p.open('r+b'))
    assert len(m) == 1
    assert m['flag'] is True
    m.close()


def test_iter(tmpdir):
    p = tmpdir.join("test.meta")

    m = Meta(p.open('w+b'))
    keys = ["foo", "bar", "baz", ]
    for key in keys:
        m[key] = True
    m.close()

    m = Meta(p.open('r+b'))
    assert set(list(m)) == set(keys)
    m.close()


def test_del(tmpdir):
    p = tmpdir.join("test.meta")
    key = "foo"

    m = Meta(p.open('w+b'))
    m[key] = True
    m.close()

    m = Meta(p.open('r+b'))
    del m[key]
    m.close()

    m = Meta(p.open('r+b'))
    assert len(m) == 0
    assert key not in m
    m.close()



from flask import render_template, url_for
from flask.views import MethodView


class QRView(MethodView):
    def get(self, name):
        target = url_for('bepasty.display', name=name, _external=True)
        return render_template('qr.html', text=target)



import importlib


def create_storage(app):
    """
    Load specified storage and return the object.
    """
    if 'STORAGE' not in app.config:
        raise Exception("Missing STORAGE config key")

    storage = importlib.import_module('.' + app.config['STORAGE'], __name__)
    return storage.create_storage(app)



from flask import Blueprint

from .delete import DeleteView
from .display import DisplayView
from .download import DownloadView, InlineView
from .modify import ModifyView
from .qr import QRView
from .filelist import FileListView
from .index import index
from .login import LoginView, LogoutView
from .setkv import LockView, UnlockView
from .upload import UploadAbortView, UploadContinueView, UploadNewView, UploadView
from .xstatic import xstatic


blueprint = Blueprint('bepasty', __name__)

blueprint.add_url_rule('/', view_func=index)
blueprint.add_url_rule('/xstatic/<name>', defaults=dict(filename=''), view_func=xstatic)
blueprint.add_url_rule('/xstatic/<name>/<path:filename>', view_func=xstatic)
blueprint.add_url_rule('/+list', view_func=FileListView.as_view('filelist'))
blueprint.add_url_rule('/+login', view_func=LoginView.as_view('login'))
blueprint.add_url_rule('/+logout', view_func=LogoutView.as_view('logout'))
blueprint.add_url_rule('/<itemname:name>', view_func=DisplayView.as_view('display'))
blueprint.add_url_rule('/<itemname:name>/+delete', view_func=DeleteView.as_view('delete'))
blueprint.add_url_rule('/<itemname:name>/+download', view_func=DownloadView.as_view('download'))
blueprint.add_url_rule('/<itemname:name>/+inline', view_func=InlineView.as_view('inline'))
blueprint.add_url_rule('/<itemname:name>/+modify', view_func=ModifyView.as_view('modify'))
blueprint.add_url_rule('/<itemname:name>/+qr', view_func=QRView.as_view('qr'))
blueprint.add_url_rule('/<itemname:name>/+lock', view_func=LockView.as_view('lock'))
blueprint.add_url_rule('/<itemname:name>/+unlock', view_func=UnlockView.as_view('unlock'))
blueprint.add_url_rule('/+upload', view_func=UploadView.as_view('upload'))
blueprint.add_url_rule('/+upload/new', view_func=UploadNewView.as_view('upload_new'))
blueprint.add_url_rule('/+upload/<itemname:name>', view_func=UploadContinueView.as_view('upload_continue'))
blueprint.add_url_rule('/+upload/<itemname:name>/abort', view_func=UploadAbortView.as_view('upload_abort'))



from hashlib import sha256 as hash_new

SIZE = 1024 * 1024


def compute_hash(data, size):
    """
    compute hash of storage item's data file, return hexdigest
    """
    hasher = hash_new()
    offset = 0
    while offset < size:
        buf = data.read(SIZE, offset)
        offset += len(buf)
        hasher.update(buf)
    return hasher.hexdigest()



"""
Set metadata keys to specific values
"""

import errno

from flask import current_app, render_template
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden

from ..constants import COMPLETE, FILENAME, LOCKED
from ..utils.http import redirect_next_referrer
from ..utils.permissions import ADMIN, may


class SetKeyValueView(MethodView):
    # overwrite these in subclasses:
    REQUIRED_PERMISSION = None
    KEY = None
    NEXT_VALUE = None

    def error(self, item, error):
        return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

    def response(self, name):
        return redirect_next_referrer('bepasty.display', name=name)

    def post(self, name):
        if self.REQUIRED_PERMISSION is not None and not may(self.REQUIRED_PERMISSION):
            raise Forbidden()
        try:
            with current_app.storage.openwrite(name) as item:
                if item.meta[self.KEY] == self.NEXT_VALUE:
                    error = '%s already is %r.' % (self.KEY, self.NEXT_VALUE)
                elif not item.meta[COMPLETE]:
                    error = 'Upload incomplete. Try again later.'
                else:
                    error = None
                if error:
                    return self.error(item, error)
                item.meta[self.KEY] = self.NEXT_VALUE
            return self.response(name)

        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise


class LockView(SetKeyValueView):
    REQUIRED_PERMISSION = ADMIN
    KEY = LOCKED
    NEXT_VALUE = True


class UnlockView(SetKeyValueView):
    REQUIRED_PERMISSION = ADMIN
    KEY = LOCKED
    NEXT_VALUE = False



import errno
import base64
import time
from io import BytesIO

from flask import Response, make_response, url_for, jsonify, stream_with_context, request, current_app
from flask.views import MethodView
from werkzeug.exceptions import HTTPException, BadRequest, Conflict, Forbidden, InternalServerError, MethodNotAllowed

from ..constants import FILENAME, ID, SIZE, TYPE, TRANSACTION_ID
from ..utils._compat import bytes_type
from ..utils.date_funcs import get_maxlife
from ..utils.http import ContentRange, DownloadRange
from ..utils.name import ItemName
from ..utils.permissions import CREATE, LIST, may
from ..utils.upload import Upload, filter_internal, background_compute_hash
from ..views.filelist import file_infos
from ..views.delete import DeleteView
from ..views.download import DownloadView
from ..views.modify import ModifyView
from ..views.setkv import LockView, UnlockView


# This wrappper handles exceptions in the REST api implementation.
#
# The @blueprint.add_errorhandler decorator could do this, but we have
# "/lodgeit/" in the same blueprint and there is no way to exclude
# from using the same error handler ("/lodgeit/" is not REST api).
def rest_errorhandler(func):
    def error_message(description, code):
        return jsonify({
            'error': {'code': code, 'message': description},
        }), code

    def handler(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPException as exc:
            return error_message(exc.description, exc.code)
        except Exception:
            if current_app.propagate_exceptions:
                # if testing/debug mode, re-raise
                raise
            exc = InternalServerError()
            return error_message(exc.description, exc.code)

    return handler


# Default handlers for REST api to handle error
class RestBase(MethodView):
    @rest_errorhandler
    def get(self, *args, **kwargs):
        raise MethodNotAllowed()

    @rest_errorhandler
    def post(self, *args, **kwargs):
        raise MethodNotAllowed()


class ItemUploadView(RestBase):
    def update_item(self, item, name):
        # Check the actual size of the file on the server against limit
        # Either 0 if new file or n bytes of already uploaded file
        Upload.filter_size(item.data.size)

        # Check Content-Range. Needs to be specified, even if only one chunk
        if not request.headers.get("Content-Range"):
            raise BadRequest(description='Content-Range not specified')

        # Get Content-Range and check if Range is consistent with server state
        file_range = ContentRange.from_request()
        if not item.data.size == file_range.begin:
            raise Conflict(description='Content-Range inconsistent. Last byte on Server: %d' % item.data.size)

        # Decode Base64 encoded request data
        try:
            raw_data = base64.b64decode(request.data)
            file_data = BytesIO(raw_data)
        except (base64.binascii.Error, TypeError):
            raise BadRequest(description='Could not decode data body')

        # Write data chunk to item
        Upload.data(item, file_data, len(raw_data), file_range.begin)

        # Make a Response and create Transaction-ID from ItemName
        response = make_response()
        response.headers['Content-Type'] = 'application/json'
        response.data = '{}'
        name_b = name if isinstance(name, bytes_type) else name.encode()
        trans_id_b = base64.b64encode(name_b)
        trans_id_s = trans_id_b if isinstance(trans_id_b, str) else trans_id_b.decode()
        response.headers[TRANSACTION_ID] = trans_id_s

        # Check if file is completely uploaded and set meta
        if file_range.is_complete:
            Upload.meta_complete(item, '')
            item.meta[SIZE] = item.data.size
            item.close()
            background_compute_hash(current_app.storage, name)
            # Set status 'successful' and return the new URL for the uploaded file
            response.status = '201'
            response.headers["Content-Location"] = url_for('bepasty_apis.items_detail', name=name)
        else:
            item.close()
            response.status = '200'

        return response

    @rest_errorhandler
    def post(self):
        """
        Upload file via REST-API. Chunked Upload is supported.

        HTTP Headers that need to be given:
        * Content-Type: The type of the file that is being uploaded.
            If this is not given filetype will be 'application/octet-stream'
        * Content-Length: The total size of the file to be uploaded.
        * Content-Filename: The filename of the file. This will be used when downloading.
        * Content-Range: The Content-Range of the Chunk that is currently being uploaded.
            Follows the HTTP-Header Specifications.
        * Transaction-ID: The Transaction-ID for Chunked Uploads.
            Needs to be delivered when uploading in chunks (after the first chunk).

        To start an upload, the HTTP Headers need to be delivered.
        The body of the request needs to be the base64 encoded file contents.
        Content-Length is the original file size before base64 encoding.
        Content-Range follows the same logic.
        After the first chunk is uploaded, bepasty will return the Transaction-ID to continue the upload.
        Deliver the Transaction-ID and the correct Content-Range to continue upload.
        After the file is completely uploaded, the file will be marked as complete and
        a 201 HTTP Status will be returned.
        The Content-Location Header will contain the api url to the uploaded Item.

        If the file size exceeds the permitted size, the upload will be aborted. This will be checked twice.
        The first check is the provided Content-Length. The second is the actual file size on the server.
        """
        if not may(CREATE):
            raise Forbidden()

        # Collect all expected data from the Request
        file_type = request.headers.get("Content-Type")
        file_size = request.headers.get("Content-Length")
        file_name = request.headers.get("Content-Filename")

        # Check the file size from Request
        Upload.filter_size(file_size)

        # Check if Transaction-ID is available for continued upload
        if not request.headers.get(TRANSACTION_ID):
            # Create ItemName and empty file in Storage
            name = ItemName.create(current_app.storage)
            item = current_app.storage.create(name, 0)

            # set max lifetime
            maxtime = get_maxlife(request.headers, underscore=False)
            maxlife_timestamp = int(time.time()) + maxtime if maxtime > 0 else maxtime
            # Fill meta with data from Request
            Upload.meta_new(item, 0, file_name, file_type,
                            'application/octet-stream',
                            name, maxlife_stamp=maxlife_timestamp)
            new_item = True
        else:
            # Get file name from Transaction-ID and open from Storage
            trans_id_s = request.headers.get(TRANSACTION_ID)
            trans_id_b = trans_id_s if isinstance(trans_id_s, bytes_type) else trans_id_s.encode()
            try:
                name_b = base64.b64decode(trans_id_b)
            except (base64.binascii.Error, TypeError):
                raise BadRequest(description='Could not decode {}'.format(TRANSACTION_ID))
            name = name_b if isinstance(name_b, str) else name_b.decode()
            try:
                item = current_app.storage.openwrite(name)
            except (OSError, IOError) as e:
                if e.errno == errno.ENOENT:
                    raise BadRequest(description='Could not find storage item for transaction id')
                raise
            new_item = False

        response = None
        try:
            response = self.update_item(item, name)
            return response
        finally:
            # If error response or exception on a new item path, remove item
            if new_item and not isinstance(response, Response):
                current_app.storage.remove(name)

    @rest_errorhandler
    def get(self):
        """
        Return the list of all files in bepasty, including metadata in the form:

            {
                "<name>": {
                    "file-meta": { <name.metadata> },
                    "uri" : "/apis/rest/items/<name>"
                },
                ...
            }
        """
        if not may(LIST):
            raise Forbidden()
        ret = {}
        for meta in file_infos():
            name = meta.pop(ID)
            ret[name] = {'uri': url_for('bepasty_apis.items_detail', name=name),
                         'file-meta': filter_internal(meta)}
        return jsonify(ret)


class ItemDetailView(DownloadView, RestBase):
    def err_incomplete(self, item, error):
        raise Conflict(description=error)

    def response(self, item, name):
        return jsonify({'uri': url_for('bepasty_apis.items_detail', name=name),
                        'file-meta': filter_internal(item.meta)})

    @rest_errorhandler
    def get(self, name):
        return super(ItemDetailView, self).get(name)


class ItemDownloadView(ItemDetailView):
    def response(self, item, name):
        request_range = DownloadRange.from_request()
        if not request_range:
            range_end = item.data.size - 1
            range_begin = 0
        else:
            if request_range.end == -1:
                range_end = item.data.size - 1
            else:
                range_end = min(request_range.end, item.data.size - 1)
            range_begin = request_range.begin

        ret = Response(stream_with_context(self.stream(item, range_begin, range_end + 1)))
        ret.headers['Content-Disposition'] = '{0}; filename="{1}"'.format(
            self.content_disposition, item.meta[FILENAME])
        ret.headers['Content-Length'] = (range_end - range_begin) + 1
        ret.headers['Content-Type'] = item.meta[TYPE]  # 'application/octet-stream'
        ret.status = '200'
        ret.headers['Content-Range'] = ('bytes %d-%d/%d' % (range_begin, range_end, item.data.size))
        return ret

    @rest_errorhandler
    def get(self, name):
        return super(ItemDetailView, self).get(name)


class ItemModifyView(ModifyView, RestBase):
    def error(self, item, error):
        raise Conflict(description=error)

    def response(self, name):
        return make_response('{}', {'Content-Type': 'application/json'})

    def get_params(self):
        json = request.json
        if json is None:
            raise BadRequest(description='Content-Type or JSON format is invalid')

        return {
            FILENAME: json.get(FILENAME),
            TYPE: json.get(TYPE),
        }

    @rest_errorhandler
    def post(self, name):
        return super(ItemModifyView, self).post(name)


class ItemDeleteView(DeleteView, RestBase):
    def error(self, item, error):
        raise Conflict(description=error)

    def response(self, name):
        return make_response('{}', {'Content-Type': 'application/json'})

    @rest_errorhandler
    def post(self, name):
        return super(ItemDeleteView, self).post(name)


class ItemLockView(LockView, RestBase):
    def error(self, item, error):
        raise Conflict(description=error)

    def response(self, name):
        return make_response('{}', {'Content-Type': 'application/json'})

    @rest_errorhandler
    def post(self, name):
        return super(ItemLockView, self).post(name)


class ItemUnlockView(UnlockView, RestBase):
    def error(self, item, error):
        raise Conflict(description=error)

    def response(self, name):
        return make_response('{}', {'Content-Type': 'application/json'})

    @rest_errorhandler
    def post(self, name):
        return super(ItemUnlockView, self).post(name)


class InfoView(RestBase):
    @rest_errorhandler
    def get(self):
        return jsonify({'MAX_BODY_SIZE': current_app.config['MAX_BODY_SIZE'],
                        'MAX_ALLOWED_FILE_SIZE': current_app.config['MAX_ALLOWED_FILE_SIZE']})



"""
bepasty-object commandline interface
"""

from __future__ import print_function

import os
import argparse
import logging
import time

from flask import Flask

from ..constants import (
    COMPLETE,
    FILENAME,
    FOREVER,
    HASH,
    LOCKED,
    SIZE,
    TIMESTAMP_DOWNLOAD,
    TIMESTAMP_MAX_LIFE,
    TIMESTAMP_UPLOAD,
    TYPE,
)
from ..utils.hashing import compute_hash
from ..storage import create_storage


class Main(object):
    argparser = argparse.ArgumentParser(prog='bepasty-object')
    _subparsers = argparser.add_subparsers()
    argparser.add_argument('--config', dest='config', metavar='CONFIG', help='bepasty configuration file')
    argparser.add_argument('names', metavar='NAME', nargs='+')

    def do_migrate(self, storage, name, args):
        tnow = time.time()
        with storage.openwrite(name) as item:
            # compatibility support for bepasty 0.0.1 and pre-0.1.0
            # old items might have a 'timestamp' value which is not used any more
            # (superseded by 'timestamp-*') - delete it:
            item.meta.pop('timestamp', None)
            # old items might miss some of the timestamps we require,
            # just initialize them with the current time:
            for ts_key in [TIMESTAMP_UPLOAD, TIMESTAMP_DOWNLOAD, ]:
                if ts_key not in item.meta:
                    item.meta[ts_key] = tnow
            if LOCKED not in item.meta:
                unlocked = item.meta.pop('unlocked', None)
                if unlocked is not None:
                    locked = not unlocked
                else:
                    locked = False
                item.meta[LOCKED] = locked
            if COMPLETE not in item.meta:
                item.meta[COMPLETE] = True
            if FILENAME not in item.meta:
                item.meta[FILENAME] = 'missing'
            if TYPE not in item.meta:
                item.meta[TYPE] = 'application/octet-stream'
            if SIZE not in item.meta:
                item.meta[SIZE] = item.data.size
            if HASH not in item.meta:
                item.meta[HASH] = ''  # see do_consistency
            if TIMESTAMP_MAX_LIFE not in item.meta:
                item.meta[TIMESTAMP_MAX_LIFE] = FOREVER

    _parser = _subparsers.add_parser('migrate', help='Migrate metadata to current schema')
    _parser.set_defaults(func=do_migrate)

    def do_purge(self, storage, name, args):
        tnow = time.time()
        with storage.openwrite(name) as item:
            file_name = item.meta[FILENAME]
            file_size = item.meta[SIZE]
            t_upload = item.meta[TIMESTAMP_UPLOAD]
            t_download = item.meta[TIMESTAMP_DOWNLOAD]
            file_type = item.meta[TYPE]
            max_lifetime = item.meta.get(TIMESTAMP_MAX_LIFE, FOREVER)
        purge = True  # be careful: we start from True, then AND the specified criteria
        if args.purge_age is not None:
            dt = args.purge_age * 24 * 3600  # n days since upload
            purge = purge and t_upload < tnow - dt
        if args.purge_inactivity is not None:
            dt = args.purge_inactivity * 24 * 3600  # n days inactivity (no download)
            purge = purge and t_download < tnow - dt
        if args.purge_size is not None:
            max_size = args.purge_size * 1024 * 1024  # size in MiB
            purge = purge and file_size > max_size
        if args.purge_type is not None:
            purge = purge and file_type.startswith(args.purge_type)
        if max_lifetime is not None:
            purge = purge and tnow > max_lifetime > 0
        if purge:
            print('removing: %s (%s %dB %s)' % (name, file_name, file_size, file_type))
            if not args.purge_dry_run:
                storage.remove(name)

    _parser = _subparsers.add_parser('purge', help='Purge objects')
    _parser.set_defaults(func=do_purge)
    _parser.add_argument('-D', '--dry-run', dest='purge_dry_run', action='store_true',
                         help='do not remove anything, just display what would happen')
    _parser.add_argument('-A', '--age', dest='purge_age', type=int, default=None,
                         help='only remove if upload older than PURGE_AGE days')
    _parser.add_argument('-I', '--inactivity', dest='purge_inactivity', type=int, default=None,
                         help='only remove if latest download older than PURGE_INACTIVITY days')
    _parser.add_argument('-S', '--size', dest='purge_size', type=int, default=None,
                         help='only remove if file size > PURGE_SIZE MiB')
    _parser.add_argument('-T', '--type', dest='purge_type', default=None,
                         help='only remove if file mimetype starts with PURGE_TYPE')

    def do_consistency(self, storage, name, args):
        with storage.openwrite(name) as item:
            file_name = item.meta[FILENAME]
            meta_size = item.meta[SIZE]
            meta_type = item.meta[TYPE]
            meta_hash = item.meta[HASH]

            print('checking: %s (%s %dB %s)' % (name, file_name, meta_size, meta_type))

            size = item.data.size
            size_consistent = size == meta_size
            if not size_consistent:
                print("Inconsistent file size: meta: %d file: %d" % (meta_size, size))
                if args.consistency_fix:
                    print("Writing computed file size into metadata...")
                    item.meta[SIZE] = size
                    size_consistent = True

            file_hash = compute_hash(item.data, size)
            hash_consistent = meta_hash == file_hash
            if not hash_consistent:
                if meta_hash:
                    print("Inconsistent hashes:")
                    print("meta: %s" % meta_hash)
                    print("file: %s" % file_hash)
                else:
                    # the upload code can not compute hashes for chunked uploads and thus stores an empty hash.
                    # we can fix that empty hash with the computed hash from the file we have in storage.
                    print("Empty hash in metadata.")
                if args.consistency_fix or args.consistency_compute and not meta_hash:
                    print("Writing computed file hash into metadata...")
                    item.meta[HASH] = file_hash
                    hash_consistent = True

        if args.consistency_remove and not (size_consistent and hash_consistent):
            print('REMOVING inconsistent file!')
            storage.remove(name)

    _parser = _subparsers.add_parser('consistency', help='Consistency-related functions')
    _parser.set_defaults(func=do_consistency)
    _parser.add_argument('-C', '--compute', dest='consistency_compute', action='store_true',
                         help='compute missing hashes and write into metadata')
    _parser.add_argument('-F', '--fix', dest='consistency_fix', action='store_true',
                         help='write computed hash/size into metadata')
    _parser.add_argument('-R', '--remove', dest='consistency_remove', action='store_true',
                         help='remove files with inconsistent hash/size')

    def do_info(self, storage, name, args):
        with storage.open(name) as item:
            print(name)
            for key, value in sorted(item.meta.items()):
                print('  ', key, value)

    _parser = _subparsers.add_parser('info', help='Display information about objects')
    _parser.set_defaults(func=do_info)

    def do_set(self, storage, name, args):
        with storage.openwrite(name) as item:
            print(name)

            if args.flag_complete is not None:
                if args.flag_complete:
                    print('  set complete')
                else:
                    print('  set not complete')
                item.meta[COMPLETE] = args.flag_complete

            if args.flag_locked is not None:
                if args.flag_locked:
                    print('  set locked')
                else:
                    print('  set not locked')
                item.meta[LOCKED] = args.flag_locked

    _parser = _subparsers.add_parser('set', help='Set flags on objects')
    _parser.set_defaults(func=do_set)
    _group = _parser.add_mutually_exclusive_group()
    _group.add_argument('-L', '--lock', dest='flag_locked', action='store_true', default=None)
    _group.add_argument('-l', '--unlock', dest='flag_locked', action='store_false', default=None)
    _group = _parser.add_mutually_exclusive_group()
    _group.add_argument('-C', '--incomplete', dest='flag_complete', action='store_false', default=None)
    _group.add_argument('-c', '--complete', dest='flag_complete', action='store_true', default=None)

    def __call__(self):
        args = Main.argparser.parse_args()

        # Setup minimal application
        app = Flask(__name__)
        app.config.from_object('bepasty.config.Config')
        if os.environ.get('BEPASTY_CONFIG'):
            app.config.from_envvar('BEPASTY_CONFIG')
        if args.config is not None:
            cfg_path = os.path.abspath(args.config)
            app.config.from_pyfile(cfg_path)
        app.storage = storage = create_storage(app)

        # Setup application context
        with app.app_context():
            # Run all before request functions by hand
            for i in app.before_request_funcs.get(None, ()):
                i()

            if len(args.names) == 1 and args.names[0] == '*':
                names = list(storage)
            else:
                names = args.names
            for name in names:
                try:
                    args.func(self, storage, name, args)
                except Exception:
                    logging.exception('Failed to handle item %s', name)


def main():
    logging.basicConfig()
    Main()()


if __name__ == '__main__':
    main()



from bepasty.storage.filesystem import Item


def test(tmpdir):
    pm = tmpdir.join("test.meta")
    pd = tmpdir.join("test.data")

    with Item(pm.open('w+b'), pd.open('w+b')) as i:
        assert i.data is not None
        assert i.meta is not None



FILENAME = 'filename'
TYPE = 'type'
TYPE_HINT = 'type-hint'
LOCKED = 'locked'
SIZE = 'size'
COMPLETE = 'complete'
HASH = 'hash'
TIMESTAMP_UPLOAD = 'timestamp-upload'
TIMESTAMP_DOWNLOAD = 'timestamp-download'
TIMESTAMP_MAX_LIFE = 'timestamp-max-life'
ID = 'id'  # storage name
FOREVER = -1

# headers
TRANSACTION_ID = 'Transaction-ID'  # keep in sync with bepasty-cli

# used internally only
internal_meta = [TYPE_HINT]



import os
import errno
from io import BytesIO
import time

from flask import abort, current_app, jsonify, request, url_for
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden
from werkzeug.urls import url_quote

from ..constants import COMPLETE, FILENAME, SIZE
from ..utils.date_funcs import get_maxlife
from ..utils.http import ContentRange, redirect_next
from ..utils.name import ItemName
from ..utils.permissions import CREATE, may
from ..utils.upload import Upload, create_item, background_compute_hash


class UploadView(MethodView):
    def post(self):
        if not may(CREATE):
            raise Forbidden()
        f = request.files.get('file')
        t = request.form.get('text')
        # note: "and f.filename" is needed due to missing __bool__ method in
        # werkzeug.datastructures.FileStorage, to work around it crashing
        # on Python 3.x.
        if f and f.filename:
            # Check Content-Range, disallow its usage
            if ContentRange.from_request():
                abort(416)

            # Check Content-Type, default to application/octet-stream
            content_type = (
                f.headers.get('Content-Type') or
                request.headers.get('Content-Type'))
            content_type_hint = 'application/octet-stream'
            filename = f.filename

            # Get size of temporary file
            f.seek(0, os.SEEK_END)
            size = f.tell()
            f.seek(0)
        elif t is not None:
            # t is already unicode, but we want utf-8 for storage
            t = t.encode('utf-8')
            content_type = request.form.get('contenttype')  # TODO: add coding
            content_type_hint = 'text/plain'
            size = len(t)
            f = BytesIO(t)
            filename = request.form.get('filename')
        else:
            raise NotImplementedError
        # set max lifetime
        maxtime = get_maxlife(request.form, underscore=False)
        maxlife_timestamp = int(time.time()) + maxtime if maxtime > 0 else maxtime
        name = create_item(f, filename, size, content_type, content_type_hint, maxlife_stamp=maxlife_timestamp)
        kw = {}
        kw['_anchor'] = url_quote(filename)
        if content_type == 'text/x-bepasty-redirect':
            # after creating a redirect, we want to stay on the bepasty
            # redirect display, so the user can copy the URL.
            kw['delay'] = '9999'
        return redirect_next('bepasty.display', name=name, **kw)


class UploadNewView(MethodView):
    def post(self):
        if not may(CREATE):
            raise Forbidden()

        data = request.get_json()

        data_filename = data['filename']
        data_size = int(data['size'])
        data_type = data['type']

        # set max lifetime
        maxtime = get_maxlife(data, underscore=True)
        maxlife_timestamp = int(time.time()) + maxtime if maxtime > 0 else maxtime

        name = ItemName.create(current_app.storage)
        with current_app.storage.create(name, data_size) as item:
            # Save meta-data
            Upload.meta_new(item, data_size, data_filename, data_type,
                            'application/octet-stream', name, maxlife_stamp=maxlife_timestamp)

            return jsonify({'url': url_for('bepasty.upload_continue', name=name),
                            'name': name})


class UploadContinueView(MethodView):
    def post(self, name):
        if not may(CREATE):
            raise Forbidden()

        f = request.files['file']
        if not f:
            raise NotImplementedError

        # Check Content-Range
        content_range = ContentRange.from_request()

        with current_app.storage.openwrite(name) as item:
            if content_range:
                # note: we ignore the hash as it is only for 1 chunk, not for the whole upload.
                # also, we can not continue computing the hash as we can't save the internal
                # state of the hash object
                size_written, _ = Upload.data(item, f, content_range.size, content_range.begin)
                file_hash = ''
                is_complete = content_range.is_complete

            else:
                # Get size of temporary file
                f.seek(0, os.SEEK_END)
                size = f.tell()
                f.seek(0)

                size_written, file_hash = Upload.data(item, f, size)
                is_complete = True

            if is_complete:
                Upload.meta_complete(item, file_hash)

            result = jsonify({'files': [{
                'name': name,
                'filename': item.meta[FILENAME],
                'size': item.meta[SIZE],
                'url': "{0}#{1}".format(url_for('bepasty.display', name=name),
                                        item.meta[FILENAME]),
            }]})

        if is_complete and not file_hash:
            background_compute_hash(current_app.storage, name)

        return result


class UploadAbortView(MethodView):
    def get(self, name):
        if not may(CREATE):
            raise Forbidden()

        try:
            item = current_app.storage.open(name)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                return 'No file found.', 404
            raise

        if item.meta[COMPLETE]:
            error = 'Upload complete. Cannot delete fileupload garbage.'
        else:
            error = None
        if error:
            return error, 409

        try:
            item = current_app.storage.remove(name)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise
        return 'Upload aborted'



from flask import request, session
from flask.views import MethodView

from ..utils.http import redirect_next_referrer
from ..utils.permissions import LOGGEDIN, PERMISSIONS, lookup_permissions


class LoginView(MethodView):
    def post(self):
        token = request.form.get('token')
        if token is not None:
            permissions_for_token = lookup_permissions(token)
            if permissions_for_token is not None:
                session[PERMISSIONS] = permissions_for_token
                session[LOGGEDIN] = True
        return redirect_next_referrer('bepasty.index')


class LogoutView(MethodView):
    def post(self):
        # note: remove all session entries that are not needed for logged-out
        # state (because the code has defaults for them if they are missing).
        # if the session is empty. flask will automatically remove the cookie.
        session.pop(LOGGEDIN, None)
        session.pop(PERMISSIONS, None)
        return redirect_next_referrer('bepasty.index')



import pytest

from bepasty.constants import FOREVER
from bepasty.utils.date_funcs import get_maxlife, time_unit_to_sec


def test_get_maxlife():
    assert get_maxlife({}, underscore=False) == 60 * 60 * 24 * 30
    assert get_maxlife({}, underscore=True) == 60 * 60 * 24 * 30


@pytest.mark.parametrize('unit,expectation', [
    ('MINUTES', 60),
    ('HOURS', 60 * 60),
    ('DAYS', 60 * 60 * 24),
    ('WEEKS', 60 * 60 * 24 * 7),
    ('MONTHS', 60 * 60 * 24 * 30),
    ('YEARS', 60 * 60 * 24 * 365),
    ('FOREVER', FOREVER),
])
def test_unit_to_secs(unit, expectation):
    assert time_unit_to_sec(1, unit) == expectation



from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import ElementNotInteractableException

import pytest
import os
import time
import tempfile


@pytest.mark.needs_server
class TestScreenShots(object):
    url_base = 'http://localhost:5000'
    # bootstrap4 breakpoints
    screenshot_dir = 'screenshots'
    screen_sizes = [(450, 700), (576, 800), (768, 600), (992, 768), (1200, 1024)]
    screenshot_seq = 1

    def setup_class(self):
        """
        Setup: Open a mozilla browser, login
        """
        self.browser = Firefox()
        self.browser.get(self.url_base + '/invalid')

    def teardown_class(self):
        """
        Tear down: Close the browser
        """
        self.browser.quit()

    def wait_present(self, xpath, timeout=10):
        cond = expected_conditions.presence_of_element_located((By.XPATH, xpath))
        WebDriverWait(self.browser, timeout).until(cond)

    def screen_shot(self, name, w, h):
        if not os.path.isdir(self.screenshot_dir):
            os.mkdir(self.screenshot_dir)
        self.browser.save_screenshot(
            '{}/{:02d}-{}-{}x{}.png'.format(self.screenshot_dir,
                                            self.screenshot_seq, name, w, h)
        )

    def screen_shots(self, name):
        for w, h in self.screen_sizes:
            self.browser.set_window_size(w, h)
            time.sleep(.1)
            self.screen_shot(name, w, h)

    def scroll_to_bottom(self):
        self.set_smallest_window_size()
        self.browser.execute_script('window.scrollTo(0, document.body.scrollHeight);')

    def set_smallest_window_size(self):
        # change smallest screen size
        w, h = self.screen_sizes[0]
        self.browser.set_window_size(w, h)
        time.sleep(.1)

    def set_largest_window_size(self):
        # change largest screen size
        w, h = self.screen_sizes[-1]
        self.browser.set_window_size(w, h)
        time.sleep(.1)

    def toggle_hamburger(self):
        self.set_smallest_window_size()

        # toggle hamburger menu
        try:
            self.browser.find_element_by_xpath('//button[@class="navbar-toggler"]').click()
        except ElementNotInteractableException:
            pass
        time.sleep(.5)

        self.set_largest_window_size()

    def top_screen_shots(self, name):
        self.screen_shots('{}1'.format(name))

        # open hamburger
        self.toggle_hamburger()

        self.screen_shots('{}2'.format(name))

        # close hamburger
        self.toggle_hamburger()

    def error_404(self):
        # NOTE: 404 error
        self.screen_shots("error404")
        self.screenshot_seq += 1

    def login(self):
        self.browser.get(self.url_base)

        # NOTE: login screen, 1 - close hamburger, 2 - open hamburger
        self.top_screen_shots("top")
        self.screenshot_seq += 1

        token = self.browser.find_element_by_name("token")
        password = "foo"
        # login
        token.send_keys(password)
        token.submit()
        self.wait_present("//input[@value='Logout']")

        # NOTE: upload screen, 1 - close hamburger, 2 - open hamburger
        self.top_screen_shots("upload")
        self.screenshot_seq += 1

        try:
            self.browser.find_element_by_xpath("//input[@value='Logout']")
        except NoSuchElementException:
            raise ValueError("Can't login! Please edit your config, go to PERMISSIONS setting "
                             "and add a new secret 'foo' with all permissions.")

    def upload_file(self, path):
        # set file path
        fileupload = self.browser.find_element_by_id('fileupload')
        fileupload.send_keys(path)

        form = self.browser.find_element_by_xpath('//form[@action="/+upload"]')
        form.click()

    def upload_view(self):
        # small files
        for i in (1, 2, 3):
            with tempfile.NamedTemporaryFile(suffix=".sh") as fp:
                fp.write("""\
#!/bin/sh

if [ $# -le 0 ]; then
    echo "no argument" 2>&1
    exit 1
fi

echo "hello, world!"
""".encode())
                fp.flush()
                self.upload_file(fp.name)

        self.scroll_to_bottom()

        # NOTE: uploaded screen
        self.screen_shots("uploading1")

        # big file
        with tempfile.NamedTemporaryFile(suffix=".bin") as fp:
            os.truncate(fp.name, 1024 * 1024 * 1024)
            self.upload_file(fp.name)

            self.scroll_to_bottom()

            # NOTE: in-progress uploading screen
            self.screen_shots("uploading2")
            self.screenshot_seq += 1

            # click abort
            abort = self.browser.find_element_by_id('fileupload-abort')
            abort.click()
            time.sleep(.5)

            # NOTE: abort bootbox
            self.screen_shots("abort")
            self.screenshot_seq += 1

            ok = self.browser.find_element_by_class_name('bootbox-accept')
            ok.click()

            self.scroll_to_bottom()

            # NOTE: aborted upload screen
            self.screen_shots("uploading3")
            self.screenshot_seq += 1

    def list_view(self):
        self.browser.get(self.url_base + '/+list')
        # NOTE: list screen
        self.screen_shots("list")
        self.screenshot_seq += 1

    def display_view(self):
        self.browser.get(self.url_base + '/+list')
        list_link = self.browser.find_elements_by_xpath('//tr/td/a')
        list_link[0].click()

        # highlight line
        self.browser.get(self.browser.current_url + '#L-4')

        # NOTE: display screen
        self.screen_shots("display")
        self.screenshot_seq += 1

        modify = self.browser.find_element_by_id('modify-btn')
        modify.click()
        time.sleep(.5)

        # NOTE: modify bootbox
        self.screen_shots("modify")
        self.screenshot_seq += 1

        modify_cancel = self.browser.find_element_by_class_name('bootbox-cancel')
        modify_cancel.click()
        time.sleep(.5)

        lock = self.browser.find_element_by_id('lock-btn')
        lock.click()
        # NOTE: display with lock screen
        self.screen_shots("lock")
        self.screenshot_seq += 1

        qr = self.browser.find_element_by_id('qr-btn')
        qr.click()
        # NOTE: QR code screen
        self.screen_shots("qr")
        self.screenshot_seq += 1

    def test(self):
        self.error_404()
        self.login()
        self.upload_view()
        self.list_view()
        self.display_view()



from xstatic.main import XStatic

# names below must be package names
mod_names = [
    'asciinema_player',
    'bootbox',
    'bootstrap',
    'font_awesome',
    'jquery',
    'jquery_ui',
    'jquery_file_upload',
    'pygments',
]

pkg = __import__('xstatic.pkg', fromlist=mod_names)
serve_files = {}
for mod_name in mod_names:
    mod = getattr(pkg, mod_name)
    xs = XStatic(mod, root_url='/static', provider='local', protocol='http')
    serve_files[xs.name] = xs.base_dir



import errno

from flask import current_app, request, render_template
from flask.views import MethodView
from werkzeug.exceptions import Forbidden, NotFound

from ..constants import COMPLETE, FILENAME, LOCKED, TYPE
from ..utils.date_funcs import delete_if_lifetime_over
from ..utils.http import redirect_next_referrer
from ..utils.permissions import ADMIN, CREATE, may
from ..utils.upload import Upload


class ModifyView(MethodView):
    def error(self, item, error):
        return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

    def response(self, name):
        return redirect_next_referrer('bepasty.display', name=name)

    def get_params(self):
        return {
            FILENAME: request.form.get('filename'),
            TYPE: request.form.get('contenttype'),
        }

    def post(self, name):
        if not may(CREATE):
            raise Forbidden()

        try:
            with current_app.storage.openwrite(name) as item:
                if not item.meta[COMPLETE] and not may(ADMIN):
                    error = 'Upload incomplete. Try again later.'
                    return self.error(item, error)

                if item.meta[LOCKED] and not may(ADMIN):
                    raise Forbidden()

                if delete_if_lifetime_over(item, name):
                    raise NotFound()

                params = self.get_params()
                if params[FILENAME]:
                    item.meta[FILENAME] = Upload.filter_filename(
                        params[FILENAME], name, params[TYPE], item.meta[TYPE]
                    )
                if params[TYPE]:
                    item.meta[TYPE], _ = Upload.filter_type(
                        params[TYPE], item.meta[TYPE]
                    )

                return self.response(name)

        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise



import errno

from flask import current_app, render_template
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden

from ..constants import COMPLETE, FILENAME, LOCKED
from ..utils.http import redirect_next_referrer
from ..utils.permissions import ADMIN, DELETE, may


class DeleteView(MethodView):
    def error(self, item, error):
        return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

    def response(self, name):
        return redirect_next_referrer('bepasty.index')

    def post(self, name):
        if not may(DELETE):
            raise Forbidden()
        try:
            with current_app.storage.open(name) as item:
                if not item.meta[COMPLETE] and not may(ADMIN):
                    error = 'Upload incomplete. Try again later.'
                    self.error(item, error)

                if item.meta[LOCKED] and not may(ADMIN):
                    raise Forbidden()

            current_app.storage.remove(name)

        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise

        return self.response(name)



import collections
from ._compat import urlparse, urljoin

from flask import request, redirect, url_for
from werkzeug.exceptions import BadRequest


# safely and comfortably redirect
# some stuff taken from / inspired by http://flask.pocoo.org/snippets/62/

def is_safe_url(target):
    """
    check if target will lead to the same server
    """
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and ref_url.netloc == test_url.netloc


def _redirect_target_url(d, use_referrer, endpoint, **values):
    """
    return redirect url to (in that order):

    - <next> from d
    - referrer (if use_referrer is True)
    - the url for endpoint/values
    """
    targets = [d.get('next'), request.referrer, url_for(endpoint, **values)]
    if not use_referrer:
        del targets[1]
    for target in targets:
        if target and is_safe_url(target):
            return target


# GET - for next 2, you may want to create urls with:
# url_for(endpoint, ..., next=something)

def get_redirect_target(endpoint, **values):
    return _redirect_target_url(request.values, False, endpoint, **values)


def get_redirect_target_referrer(endpoint, **values):
    return _redirect_target_url(request.values, True, endpoint, **values)


# POST - for next 2, you may want this in the form:
# <input type=hidden name="next" value="{{ next or '' }}">

def redirect_next(endpoint, **values):
    return redirect(_redirect_target_url(request.form, False, endpoint, **values))


def redirect_next_referrer(endpoint, **values):
    return redirect(_redirect_target_url(request.form, True, endpoint, **values))


class ContentRange(collections.namedtuple('ContentRange', ('begin', 'end', 'complete'))):
    """
    Work with Content-Range headers.
    """
    __slots__ = ()

    @classmethod
    def parse(cls, content_range):
        """
        Parse Content-Range header.
        Format is "bytes 0-524287/2000000".
        """
        try:
            range_type, range_count = content_range.split(' ', 1)
        except ValueError:
            raise BadRequest(description='Content-Range Header is incorrect')
        # There are no other types then "bytes"
        if range_type != 'bytes':
            raise BadRequest

        try:
            range_count, range_complete = range_count.split('/', 1)
        except ValueError:
            raise BadRequest(description='Content-Range Header is incorrect')

        try:
            # For now, */2000000 format is not supported
            range_begin, range_end = range_count.split('-', 1)

            range_begin = int(range_begin)
            range_end = int(range_end)
        except ValueError:
            raise BadRequest(description='Content-Range Header has invalid range')

        # For now, 0-10/* format is not supported
        try:
            range_complete = int(range_complete)
        except ValueError:
            raise BadRequest(description='Content-Range Header has invalid length')

        if range_begin <= range_end < range_complete:
            return ContentRange(range_begin, range_end, range_complete)

        raise BadRequest

    @classmethod
    def from_request(cls):
        """
        Read Content-Range from request and parse it
        """
        content_range = request.headers.get('Content-Range')
        if content_range is not None:
            return cls.parse(content_range)

    @property
    def is_complete(self):
        return self.complete == self.end + 1

    @property
    def size(self):
        return self.end - self.begin + 1


class DownloadRange(collections.namedtuple('DownloadRange', ('begin', 'end'))):
    """
    Work with Range headers.
    """
    __slots__ = ()

    @classmethod
    def parse(cls, content_range):
        """
        Parse Range header.
        Format is "bytes=0-524287".
        """
        try:
            range_type, range_count = content_range.split('=', 1)
        except ValueError:
            raise BadRequest(description='Range Header is incorrect')
        # There are no other types than "bytes"
        if range_type != 'bytes':
            raise BadRequest(description='Range Header is incorrect')

        try:
            range_begin, range_end = range_count.split('-', 1)
        except ValueError:
            raise BadRequest(description='Range Header is incorrect')

        try:
            range_begin = int(range_begin)
        except ValueError:
            raise BadRequest(description='Range Header has invalid first')

        if range_end:
            # For now, set of ranges (e.g. 0-1,2-10) is not supported
            try:
                range_end = int(range_end)
            except ValueError:
                raise BadRequest(description='Range Header has invalid last')
        else:
            range_end = -1

        if range_begin <= range_end or range_end == -1:
            return DownloadRange(range_begin, range_end)

        raise BadRequest(description='Range Header is incorrect')

    @classmethod
    def from_request(cls):
        """
        Read Content-Range from request and parse it
        """
        download_range = request.headers.get('Range')
        if download_range is not None:
            return cls.parse(download_range)

    @property
    def size(self):
        return self.end - self.begin + 1



import re
import time
import mimetypes
from werkzeug.exceptions import BadRequest, RequestEntityTooLarge

from flask import current_app

try:
    import magic as magic_module
    magic = magic_module.Magic(mime=True)
    magic_bufsz = magic.getparam(magic_module.MAGIC_PARAM_BYTES_MAX)
except ImportError:
    magic = None
    magic_bufsz = None

from ..constants import (
    COMPLETE,
    FILENAME,
    FOREVER,
    HASH,
    LOCKED,
    SIZE,
    TIMESTAMP_DOWNLOAD,
    TIMESTAMP_MAX_LIFE,
    TIMESTAMP_UPLOAD,
    TYPE,
    TYPE_HINT,
    internal_meta,
)
from .name import ItemName
from .decorators import threaded
from .hashing import compute_hash, hash_new

# we limit to 250 characters as we do not want to accept arbitrarily long
# filenames. other than that, there is no specific reason we could not
# also take more (or less).
MAX_FILENAME_LENGTH = 250


class Upload(object):
    _filename_re = re.compile(r'[^a-zA-Z0-9 *+:;.,_-]+')
    _type_re = re.compile(r'[^a-zA-Z0-9/+.-]+')

    @classmethod
    def filter_size(cls, i):
        """
        Filter size.
        Check for advertised size.
        """
        try:
            i = int(i)
        except (ValueError, TypeError):
            raise BadRequest(description='Size is invalid')
        if i > current_app.config['MAX_ALLOWED_FILE_SIZE']:
            raise RequestEntityTooLarge()
        return i

    @classmethod
    def filter_filename(cls, filename, storage_name, content_type, content_type_hint):
        """
        Filter filename.
        Only allow some basic characters and shorten to 50 characters.
        """
        # Make up filename if we don't have one
        if not filename:
            if not content_type:
                content_type = content_type_hint
            # note: stdlib mimetypes.guess_extension is total crap
            if content_type.startswith("text/"):
                ext = ".txt"
            else:
                ext = ".bin"
            filename = storage_name + ext
        return cls._filename_re.sub('', filename)[:MAX_FILENAME_LENGTH]

    @classmethod
    def filter_type(cls, ct, ct_hint, filename=None):
        """
        Filter Content-Type
        Only allow some basic characters and shorten to 50 characters.

        Return value:
        tuple[0] - content-type string
        tuple[1] - whether tuple[0] is hint or not
                   True:  content-type is just a hint
                   False: content-type is not a hint, was specified by user
        """
        if not ct and filename:
            ct, encoding = mimetypes.guess_type(filename)
        if not ct:
            return ct_hint, True
        return cls._type_re.sub('', ct)[:50], False

    @classmethod
    def meta_new(cls, item, input_size, input_filename, input_type,
                 input_type_hint, storage_name, maxlife_stamp=FOREVER):
        item.meta[FILENAME] = cls.filter_filename(
            input_filename, storage_name, input_type, input_type_hint
        )
        item.meta[SIZE] = cls.filter_size(input_size)
        ct, hint = cls.filter_type(input_type, input_type_hint, input_filename)
        item.meta[TYPE] = ct
        item.meta[TYPE_HINT] = hint
        item.meta[TIMESTAMP_UPLOAD] = int(time.time())
        item.meta[TIMESTAMP_DOWNLOAD] = 0
        item.meta[LOCKED] = current_app.config['UPLOAD_LOCKED']
        item.meta[COMPLETE] = False
        item.meta[HASH] = ''
        item.meta[TIMESTAMP_MAX_LIFE] = maxlife_stamp

    @classmethod
    def meta_complete(cls, item, file_hash):
        # update TYPE by python-magic if not decided yet
        if item.meta.pop(TYPE_HINT, False):
            if magic and current_app.config.get('USE_PYTHON_MAGIC', False):
                if item.meta[TYPE] == 'application/octet-stream':
                    item.meta[TYPE] = magic.from_buffer(item.data.read(magic_bufsz, 0))
        item.meta[COMPLETE] = True
        item.meta[HASH] = file_hash

    @staticmethod
    def data(item, f, size_input, offset=0):
        """
        Copy data from temp file into storage.
        """
        read_length = 16 * 1024
        size_written = 0
        hasher = hash_new()

        while True:
            read_length = min(read_length, size_input)
            if size_input == 0:
                break

            buf = f.read(read_length)
            if not buf:
                # Should not happen, we already checked the size
                raise RuntimeError

            item.data.write(buf, offset + size_written)
            hasher.update(buf)

            len_buf = len(buf)
            size_written += len_buf
            size_input -= len_buf

        return size_written, hasher.hexdigest()


def create_item(f, filename, size, content_type, content_type_hint,
                maxlife_stamp=FOREVER):
    """
    create an item from open file <f> with the given metadata, return the item name.
    """
    name = ItemName.create(current_app.storage)
    with current_app.storage.create(name, size) as item:
        size_written, file_hash = Upload.data(item, f, size)
        Upload.meta_new(item, size, filename, content_type, content_type_hint,
                        name, maxlife_stamp=maxlife_stamp)
        Upload.meta_complete(item, file_hash)
    return name


def filter_internal(meta):
    """
    filter internal meta data out.
    """
    return {k: v for k, v in meta.items() if k not in internal_meta}


@threaded
def background_compute_hash(storage, name):
    with storage.openwrite(name) as item:
        size = item.meta[SIZE]
        file_hash = compute_hash(item.data, size)
        item.meta[HASH] = file_hash



from io import BytesIO

from flask import request
from flask.views import MethodView
from pygments.lexers import get_all_lexers
from werkzeug.exceptions import Forbidden
from werkzeug.urls import url_quote

from ..constants import FOREVER
from ..utils.http import redirect_next
from ..utils.permissions import CREATE, may
from ..utils.upload import create_item


class LodgeitUpload(MethodView):
    """
    lodgeit paste form
    """
    # most stuff lodgeit support comes directly from pygments
    # for all other stuff we fall back to text/plain.
    TRANS = {}
    for lexer in get_all_lexers():
        # (name, aliases, filetypes, mimetypes)
        # e.g. ('Diff', ('diff',), ('*.diff', '*.patch'), ('text/x-diff', 'text/x-patch'))
        name = lexer[1][0]
        cts = lexer[3]
        # find a content-type, preferably one with text/*
        for ct in cts:
            if ct.startswith("text/"):
                break
        else:
            if cts:
                ct = cts[0]
            else:
                ct = None
        if ct:
            TRANS[name] = ct

    def post(self):
        if not may(CREATE):
            raise Forbidden()
        lang = request.form.get('language')
        content_type = self.TRANS.get(lang)
        content_type_hint = 'text/plain'
        filename = None
        t = request.form['code']
        # t is already unicode, but we want utf-8 for storage
        t = t.encode('utf-8')
        size = len(t)
        f = BytesIO(t)
        maxlife_timestamp = FOREVER
        name = create_item(f, filename, size, content_type, content_type_hint,
                           maxlife_stamp=maxlife_timestamp)
        return redirect_next('bepasty.display', name=name, _anchor=url_quote(filename))



import os
import pickle
import logging
import tempfile

from ...utils._compat import MutableMapping

logger = logging.getLogger(__name__)


def create_storage(app):
    # decouple Storage class from flask app
    storage_dir = app.config['STORAGE_FILESYSTEM_DIRECTORY']
    return Storage(storage_dir)


class Storage(object):
    """
    Filesystem storage - store meta and data into separate files in a directory.
    """
    def __init__(self, storage_dir):
        try:
            fd, fname = tempfile.mkstemp(dir=storage_dir)
        except OSError as e:
            logger.error("Could not write file in storage directory: %s\n %s", storage_dir, e)
            raise
        else:
            os.close(fd)
            os.remove(fname)
            self.directory = storage_dir

    def _filename(self, name):
        if '/' in name:
            raise RuntimeError
        return os.path.join(self.directory, name)

    def _open(self, name, mode):
        basefilename = self._filename(name)
        file_data = open(basefilename + '.data', mode)
        file_meta = open(basefilename + '.meta', mode)
        return Item(file_data, file_meta)

    def create(self, name, size):
        return self._open(name, 'w+b')

    def open(self, name):
        return self._open(name, 'rb')

    def openwrite(self, name):
        return self._open(name, 'r+b')

    def remove(self, name):
        basefilename = self._filename(name)
        file_data = basefilename + '.data'
        file_meta = basefilename + '.meta'
        try:
            os.remove(file_data)
        except OSError as e:
            logger.error("Could not delete file: %s\n %s" % (file_data, str(e)))
            raise
        try:
            os.remove(file_meta)
        except OSError as e:
            logger.error("Could not delete file: %s\n %s" % (file_meta, str(e)))
            raise

    def __iter__(self):
        names = [fn[:-5] for fn in os.listdir(self.directory)
                 if fn.endswith('.meta')]
        for name in names:
            yield name

    def __contains__(self, name):
        meta_filename = self._filename(name) + '.meta'
        return os.path.exists(meta_filename)


class Item(object):
    """
    Represents an open item.

    :ivar data: Open file-like object to data.
    """
    def __init__(self, file_data, file_meta):
        """
        :param file_data: Open file-like object to data file.
        :param file_meta: Open file-like object to meta file.
        """
        self.data = Data(file_data)
        self.meta = Meta(file_meta)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.data.close()
        self.meta.close()

    close = __exit__


class Data(object):
    """
    Data of item.
    """
    def __init__(self, file_data):
        self._file = file_data

    @property
    def size(self):
        self._file.seek(0, os.SEEK_END)
        return self._file.tell()

    def close(self):
        self._file.close()

    def read(self, size, offset):
        self._file.seek(offset)
        return self._file.read(size)

    def write(self, data, offset):
        self._file.seek(offset)
        return self._file.write(data)


class Meta(MutableMapping):
    """
    Meta-data of item.
    """
    def __init__(self, file_meta):
        self._changed = False
        self._file = file_meta
        data = file_meta.read()
        if data:
            self._data = pickle.loads(data)
        else:
            # for empty input, we create a usable, empty Meta item,
            # but we do NOT set self._changed to True to avoid trying
            # to write to files opened in read-only mode.
            self._data = {}

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self._changed = True

    def __delitem__(self, key):
        del self._data[key]
        self._changed = True

    def close(self):
        self.write()
        self._file.close()

    def write(self):
        if self._changed:
            self._write()
            self._changed = False

    def _write(self):
        self._file.seek(0)
        # Python 2.x only uses protocol 2, 3.x uses protocol 3 by default.
        # If we just use protocol 2, changing the Python version shouldn't
        # cause problems with existing pickles.
        pickle.dump(self._data, self._file, protocol=2)
        self._file.seek(0)



import re
import random

from werkzeug.routing import BaseConverter

ID_LENGTH = 8

letters_lower = set("abcdefghijklmnopqrstuvwxyz")
letters_upper = set("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
digits = set("0123456789")

# this stuff might be hard to read / differentiate or is otherwise unwanted:
unwanted = set(
    "1lI"
    "0O"
    "8B"
    "5S"
    "+"  # used in URL dispatching for views, e.g. +login
    "/"  # used in URLs and fs paths
)

all_chars = letters_lower | letters_upper | digits
all_chars -= unwanted
all_chars = ''.join(sorted(all_chars))


def encode(x, length, alphabet=all_chars):
    """
    encode x in alphabet (with "leading zeroes")

    :param x: integer number
    :param length: length of return sequence
    :param alphabet: alphabet to choose characters from (default: all_chars)
    :return: sequence of alphabet members [list]
    """
    if x < 0:
        raise ValueError("negative values are not supported")
    if length < 0:
        raise ValueError("length must be >= 0")
    n = len(alphabet)
    if n < 2:
        raise ValueError("alphabet must be at least 2 elements long")
    code = []
    counter = length
    while x > 0 and counter > 0:
        x, r = divmod(x, n)
        code.append(alphabet[r])
        counter -= 1
    leading_zeroes = length - len(code)
    code += list(alphabet[0] * leading_zeroes)
    return list(reversed(code))


def make_id(length, x=None, alphabet=all_chars):
    """
    generate a id of <length> from value <x>.
    if x is None, use a random value for x.
    for the id, use elements from alphabet.
    """
    base = len(alphabet)  # e.g. 10
    if x is None:
        x = random.randint(
            0,                  # 000 (length==3) ...
            base ** length - 1  # 999 (length==3)
        )
    return ''.join(encode(x, length, alphabet))


class ItemName(str):
    def __new__(cls, uid):
        return str(uid)

    @classmethod
    def create(cls, storage, length=ID_LENGTH, max_length=2 * ID_LENGTH, max_tries=10):
        """
        create a unique item name in storage, wanted name length is <length>.

        we try at most <max_tries> times to find a unique name of a specific length -
        if we do not succeed, we increase name length and try again.
        if we can't find a unique name even for longer lengths up to max_length,
        we'll raise RuntimeError.
        """
        name = None  # avoid false alarm about reference before assignment
        while length <= max_length:
            tries = 0
            while tries < max_tries:
                name = make_id(length)
                if name not in storage:
                    break
                tries += 1
            if tries < max_tries:
                # we found a name, break out of outer while also
                break
            length += 1
        if length > max_length:
            raise RuntimeError("no unique names available")
        return cls(name)


class ItemNameConverter(BaseConverter):
    """
    Accept the names of the style as we generate.
    """
    # for a while, support both old uuid4-style as well as new shorter IDs
    regex = (
        '([a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12})'
        '|'
        '([%(valid)s]{%(length)d})'
    ) % dict(valid=re.escape(all_chars), length=ID_LENGTH)
    weight = 200


def setup_werkzeug_routing(app):
    app.url_map.converters['itemname'] = ItemNameConverter



import errno
import time

from flask import Response, current_app, render_template, stream_with_context
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden

from ..constants import COMPLETE, FILENAME, LOCKED, SIZE, TIMESTAMP_DOWNLOAD, TYPE
from ..utils.date_funcs import delete_if_lifetime_over
from ..utils.permissions import ADMIN, READ, may


class DownloadView(MethodView):
    content_disposition = 'attachment'  # to trigger download

    def err_incomplete(self, item, error):
        return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

    def stream(self, item, start, limit):
        with item as _item:
            # Stream content from storage
            offset = max(0, start)
            while offset < limit:
                buf = _item.data.read(min(limit - offset, 16 * 1024), offset)
                offset += len(buf)
                yield buf
            item.meta[TIMESTAMP_DOWNLOAD] = int(time.time())

    def response(self, item, name):
        ct = item.meta[TYPE]
        dispo = self.content_disposition
        if dispo != 'attachment':
            # no simple download, so we must be careful about XSS
            if ct.startswith("text/"):
                ct = 'text/plain'  # only send simple plain text

        ret = Response(stream_with_context(self.stream(item, 0, item.data.size)))
        ret.headers['Content-Disposition'] = '{0}; filename="{1}"'.format(
            dispo, item.meta[FILENAME])
        ret.headers['Content-Length'] = item.meta[SIZE]
        ret.headers['Content-Type'] = ct
        ret.headers['X-Content-Type-Options'] = 'nosniff'  # yes, we really mean it
        return ret

    def get(self, name):
        if not may(READ):
            raise Forbidden()
        try:
            item = current_app.storage.openwrite(name)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise

        try:
            need_close = True
            if not item.meta[COMPLETE]:
                return self.err_incomplete(item, 'Upload incomplete. Try again later.')

            if item.meta[LOCKED] and not may(ADMIN):
                raise Forbidden()

            if delete_if_lifetime_over(item, name):
                raise NotFound()
            need_close = False
        finally:
            if need_close:
                item.close()

        return self.response(item, name)


class InlineView(DownloadView):
    content_disposition = 'inline'  # to trigger viewing in browser, for some types



from flask import Blueprint

from .lodgeit import LodgeitUpload
from .rest import ItemDetailView, ItemDownloadView, ItemModifyView, \
    ItemUploadView, InfoView, ItemDeleteView, ItemLockView, ItemUnlockView


blueprint = Blueprint('bepasty_apis', __name__, url_prefix='/apis')

blueprint.add_url_rule('/lodgeit/', view_func=LodgeitUpload.as_view('lodgeit'))
blueprint.add_url_rule('/rest', view_func=InfoView.as_view('api_info'))
blueprint.add_url_rule('/rest/items', view_func=ItemUploadView.as_view('items'))
blueprint.add_url_rule('/rest/items/<itemname:name>', view_func=ItemDetailView.as_view('items_detail'))
blueprint.add_url_rule('/rest/items/<itemname:name>/download', view_func=ItemDownloadView.as_view('items_download'))
blueprint.add_url_rule('/rest/items/<itemname:name>/delete', view_func=ItemDeleteView.as_view('items_delete'))
blueprint.add_url_rule('/rest/items/<itemname:name>/modify', view_func=ItemModifyView.as_view('items_modify'))
blueprint.add_url_rule('/rest/items/<itemname:name>/lock', view_func=ItemLockView.as_view('items_lock'))
blueprint.add_url_rule('/rest/items/<itemname:name>/unlock', view_func=ItemUnlockView.as_view('items_unlock'))



import time
from flask import current_app
from werkzeug.exceptions import BadRequest

from ..constants import FOREVER, TIMESTAMP_MAX_LIFE


def get_maxlife(data, underscore):
    unit_key = 'maxlife_unit' if underscore else 'maxlife-unit'
    unit_default = 'MONTHS'
    unit = data.get(unit_key, unit_default).upper()
    value_key = 'maxlife_value' if underscore else 'maxlife-value'
    value_default = '1'
    try:
        value = int(data.get(value_key, value_default))
    except (ValueError, TypeError):
        raise BadRequest(description='{} header is incorrect'.format(value_key))
    try:
        return time_unit_to_sec(value, unit)
    except KeyError:
        raise BadRequest(description='{} header is incorrect'.format(unit_key))


def time_unit_to_sec(value, unit):
    """
    Converts a numeric value and with a string time unit unit to a time in seconds

    :param value: int
    :param unit: str in ['MINUTES', 'HOURS', 'DAYS', 'WEEKS', 'MONTHS', 'YEARS', 'FOREVER']
    :return: time in seconds
    """
    units = {
        'MINUTES': 60,
        'HOURS': 60 * 60,
        'DAYS': 60 * 60 * 24,
        'WEEKS': 60 * 60 * 24 * 7,
        'MONTHS': 60 * 60 * 24 * 30,
        'YEARS': 60 * 60 * 24 * 365,
        'FOREVER': FOREVER,
    }
    secs = units[unit] * value if units[unit] > 0 else units[unit]
    return secs


def delete_if_lifetime_over(item, name):
    """
    :return: True if file was deleted
    """
    if 0 < item.meta[TIMESTAMP_MAX_LIFE] < time.time():
        try:
            current_app.storage.remove(name)
        except (OSError, IOError):
            pass
        return True
    return False



from threading import Thread


def threaded(func):
    """
    decorator to run a function asynchronously (in a thread)

    be careful: do not access flask threadlocals in f!
    """
    def wrapper(*args, **kwargs):
        t = Thread(target=func, args=args, kwargs=kwargs)
        t.start()
    return wrapper



import pytest

from bepasty.storage.filesystem import Storage


def test_contains(tmpdir):
    storage = Storage(str(tmpdir))
    name = "foo"
    # check if it is not there yet
    assert name not in storage
    with storage.create(name, 0):
        # we just want it created, no need to write sth into it
        pass
    # check if it is there
    assert name in storage
    storage.remove(name)
    # check if it is gone
    assert name not in storage


def test_iter(tmpdir):
    storage = Storage(str(tmpdir))
    # nothing there yet
    assert list(storage) == []
    names = ["foo", "bar", "baz", ]
    for name in names:
        with storage.create(name, 0):
            # we just want it created, no need to write sth into it
            pass
    assert set(list(storage)) == set(names)


def test_invalid_name(tmpdir):
    storage = Storage(str(tmpdir))
    name = "../invalid"
    with pytest.raises(RuntimeError):
        storage.create(name, 0)



#
# app tests
#

from flask import request, url_for
from flask.views import MethodView

from ..app import create_app
from ..config import Config


def test_secret_key(monkeypatch):
    monkeypatch.setattr(Config, 'PERMISSIONS', {
        'admin': 'admin,list,create,read,delete',
        'full': 'list,create,read,delete',
        'none': '',
    })
    monkeypatch.setattr(Config, 'SECRET_KEY', 'secret')

    app = create_app()
    secret_key = app.config['SECRET_KEY']
    assert len(secret_key) > len(Config.SECRET_KEY)

    Config.PERMISSIONS = {
        'admin': 'admin,list,create,read,delete',
        'none': '',
    }
    app = create_app()
    assert app.config['SECRET_KEY'] != secret_key


class TestView(MethodView):
    callback = None

    def get(self):
        TestView.callback()
        return 'done'


def prepare(callback):
    app = create_app()
    app.add_url_rule('/test_call', view_func=TestView.as_view('test.test_call'))

    TestView.callback = staticmethod(callback)
    client = app.test_client()

    assert app.config['APP_BASE_PATH'] == Config.APP_BASE_PATH

    return app, client


def test_none(monkeypatch):
    monkeypatch.setattr(Config, 'APP_BASE_PATH', None)

    def none_callback():
        url = url_for('test.test_call')
        assert url == request.path

    app, client = prepare(none_callback)

    response = client.get('/bepasty/test_call')
    assert response.status_code == 404

    response = client.get('/test_call')
    assert response.status_code == 200
    assert response.data == b'done'


def test_prefix(monkeypatch):
    monkeypatch.setattr(Config, 'APP_BASE_PATH', '/bepasty')

    def prefix_callback():
        url = url_for('test.test_call')
        assert url == Config.APP_BASE_PATH + request.path

    app, client = prepare(prefix_callback)

    response = client.get('/test_call')
    assert response.status_code == 404

    response = client.get('/bepasty/test_call')
    assert response.status_code == 200
    assert response.data == b'done'



from os import close, unlink
from random import random
from tempfile import mkstemp

import pytest

from bepasty.app import create_app


@pytest.yield_fixture(scope='module')
def app(request):
    """
    creates a bepasty App-Instance
    """
    app = create_app()
    yield app
    unlink(app.config['DATABASE'])


@pytest.yield_fixture(scope='module')
def testclient(request, app):
    """
    creates a Flask-testclient instance for bepasty
    """
    db_file, app.config['DATABASE'] = mkstemp()
    # reset default permissions
    app.config['DEFAULT_PERMISSIONS'] = ''
    # setup a secret key
    app.config['SECRET_KEY'] = str(random())
    # setup permissions
    app.config['PERMISSIONS'] = {
        'l': 'list',
        'c': 'create',
        'r': 'read',
        'd': 'delete',
        'a': 'admin'
    }
    yield app.test_client()
    close(db_file)



from flask import send_from_directory
from werkzeug.exceptions import NotFound

from ..bepasty_xstatic import serve_files


def xstatic(name, filename):
    """Route to serve the xstatic files (from serve_files)"""
    try:
        base_path = serve_files[name]
    except KeyError:
        raise NotFound()

    if not filename:
        raise NotFound()

    return send_from_directory(base_path, filename)



from flask import render_template

from pygments.lexers import get_all_lexers


def contenttypes_list():
    contenttypes = [
        'text/x-bepasty-redirect',  # redirect / link shortener service
    ]
    for lexer_info in get_all_lexers():
        contenttypes.extend(lexer_info[3])
    return contenttypes


def index():
    return render_template('index.html', contenttypes=contenttypes_list())



import sys

__all__ = ['urljoin', 'urlparse']

PY2 = sys.version_info[0] == 2

if PY2:
    from urlparse import urlparse, urljoin

    range_type = xrange  # noqa: F821
    bytes_type = str
    iteritems = lambda d: d.iteritems()  # noqa: E731
    from collections import MutableMapping  # noqa: F401
else:
    from urllib.parse import urlparse, urljoin

    range_type = range
    bytes_type = bytes
    iteritems = lambda d: iter(d.items())  # noqa: E731
    from collections.abc import MutableMapping  # noqa: F401



import errno
import time

from flask import current_app, render_template, Markup, url_for, request
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden
from pygments import highlight
from pygments.lexers import get_lexer_for_mimetype
from pygments.util import ClassNotFound as NoPygmentsLexer

from ..constants import COMPLETE, FILENAME, LOCKED, SIZE, TIMESTAMP_DOWNLOAD, TYPE
from ..utils._compat import iteritems
from ..utils.date_funcs import delete_if_lifetime_over
from ..utils.formatters import CustomHtmlFormatter
from ..utils.permissions import ADMIN, READ, may

from .index import contenttypes_list
from .filelist import file_infos


def rendering_allowed(item_type, item_size, use_pygments, complete):
    """
    check if rendering is allowed, checks for:

    * whether the item is completely uploaded
    * whether the size is within the configured limits for the content-type
    """
    if not complete:
        return False
    if use_pygments:
        # if we use pygments, special restrictions apply
        item_type = 'HIGHLIGHT_TYPES'
    # create a tuple list [(content_type_prefix, max_size), ...] with long prefixes first
    ct_size = sorted(iteritems(current_app.config['MAX_RENDER_SIZE']), key=lambda e: len(e[0]), reverse=True)
    for ct, size in ct_size:
        if item_type.startswith(ct):
            return item_size <= size
    # there should be one entry with ct == '', so we should never get here:
    return False


class DisplayView(MethodView):
    def get(self, name):
        if not may(READ):
            raise Forbidden()
        try:
            item = current_app.storage.openwrite(name)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise

        with item as item:
            complete = item.meta[COMPLETE]
            if not complete and not may(ADMIN):
                error = 'Upload incomplete. Try again later.'
                return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

            if item.meta[LOCKED] and not may(ADMIN):
                raise Forbidden()

            if delete_if_lifetime_over(item, name):
                raise NotFound()

            def read_data(item):
                # reading the item for rendering is registered like a download
                data = item.data.read(item.data.size, 0)
                item.meta[TIMESTAMP_DOWNLOAD] = int(time.time())
                return data

            size = item.meta[SIZE]
            ct = item.meta[TYPE]
            try:
                get_lexer_for_mimetype(ct)
                use_pygments = True
                ct_pygments = ct
            except NoPygmentsLexer:
                if ct.startswith('text/'):
                    # seems like we found a text type not supported by pygments
                    # use text/plain so we get a display with line numbers
                    use_pygments = True
                    ct_pygments = 'text/plain'
                else:
                    use_pygments = False

            if rendering_allowed(ct, size, use_pygments, complete):
                if ct.startswith('text/x-bepasty-'):
                    # special bepasty items - must be first, don't feed to pygments
                    if ct == 'text/x-bepasty-list':
                        names = read_data(item).decode('utf-8').splitlines()
                        files = sorted(file_infos(names), key=lambda f: f[FILENAME])
                        rendered_content = Markup(render_template('filelist_tableonly.html', files=files))
                    elif ct == 'text/x-bepasty-redirect':
                        url = read_data(item).decode('utf-8')
                        delay = request.args.get('delay', '3')
                        return render_template('redirect.html', url=url, delay=delay)
                    else:
                        rendered_content = u"Can't render this content type."
                elif ct.startswith('image/'):
                    src = url_for('bepasty.download', name=name)
                    rendered_content = Markup(u'<img src="%s" alt="the image" width="800">' % src)
                elif ct.startswith('audio/'):
                    src = url_for('bepasty.download', name=name)
                    alt_msg = u'html5 audio element not supported by your browser.'
                    rendered_content = Markup(u'<audio controls src="%s">%s</audio>' % (src, alt_msg))
                elif ct.startswith('video/'):
                    src = url_for('bepasty.download', name=name)
                    alt_msg = u'html5 video element not supported by your browser.'
                    rendered_content = Markup(u'<video controls src="%s">%s</video>' % (src, alt_msg))
                elif ct in ['application/pdf', 'application/x-pdf', ]:
                    src = url_for('bepasty.inline', name=name)
                    link_txt = u'Click to see PDF'
                    rendered_content = Markup(u'<a href="%s">%s</a>' % (src, link_txt))
                elif ct == 'application/x-asciinema-recording':
                    src = url_for('bepasty.download', name=name)
                    rendered_content = Markup(u'<asciinema-player src="%s" title="%s" theme="%s" autoplay>' %
                                              (src, item.meta[FILENAME],
                                               current_app.config.get('ASCIINEMA_THEME', 'asciinema')))
                elif use_pygments:
                    text = read_data(item)
                    # TODO we don't have the coding in metadata
                    try:
                        text = text.decode('utf-8')
                    except UnicodeDecodeError:
                        # well, it is not utf-8 or ascii, so we can only guess...
                        text = text.decode('iso-8859-1')
                    lexer = get_lexer_for_mimetype(ct_pygments)
                    formatter = CustomHtmlFormatter(linenos='table', lineanchors="L",
                                                    lineparagraphs="L", anchorlinenos=True)
                    rendered_content = Markup(highlight(text, lexer, formatter))
                else:
                    rendered_content = u"Can't render this content type."
            else:
                if not complete:
                    rendered_content = u"Rendering not allowed (not complete). Is it still being uploaded?"
                else:
                    rendered_content = u"Rendering not allowed (too big?). Try download"

            return render_template('display.html', name=name, item=item,
                                   rendered_content=rendered_content,
                                   contenttypes=contenttypes_list())



"""
bepasty-server commandline interface
"""

from __future__ import print_function
import argparse

from ..app import create_app


def main():
    argparser = argparse.ArgumentParser(prog='bepasty-server')
    argparser.add_argument('--host', help='Host to listen on')
    argparser.add_argument('--port', type=int, help='Port to listen on')
    argparser.add_argument('--debug', help='Activate debug mode', action='store_true')

    args = argparser.parse_args()
    app = create_app()
    print(" * Starting bepasty server...")
    app.run(host=args.host, port=args.port, debug=args.debug)


if __name__ == '__main__':
    main()



from flask import request, session, current_app
from flask import g as flaskg


# in the code, please always use this constants for permission values:
ADMIN = 'admin'
LIST = 'list'
CREATE = 'create'
MODIFY = 'modify'
READ = 'read'
DELETE = 'delete'

# key in the session:
PERMISSIONS = 'permissions'
LOGGEDIN = 'loggedin'

permission_icons = {
    'admin': 'user',
    'list': 'list',
    'create': 'plus',
    'modify': 'edit',
    'read': 'book',
    'delete': 'trash'
}


def lookup_permissions(token):
    """
    look up the permissions string for the secret <token> in the configuration.
    if no such secret is configured, return None
    """
    return current_app.config['PERMISSIONS'].get(token)


def get_permissions():
    """
    get the permissions for the current user (if logged in)
    or the default permissions (if not logged in).
    """
    auth = request.authorization
    if auth:
        # http basic auth header present
        permissions = lookup_permissions(auth.password)
    elif 'token' in request.values:
        # token present in query args or post form (can be used by cli clients)
        permissions = lookup_permissions(request.values['token'])
    else:
        # look into session, login might have put something there
        permissions = session.get(PERMISSIONS)
    if permissions is None:
        permissions = current_app.config['DEFAULT_PERMISSIONS']
    permissions = set(permissions.split(','))
    return permissions


def get_permission_icons():
    return [
        (permission, permission_icons[permission])
        for permission in sorted(get_permissions()) if permission
    ]


def may(permission):
    """
    check whether the current user has the permission <permission>
    """
    return permission in flaskg.permissions


def logged_in():
    """
    is the user logged-in?
    """
    return session.get(LOGGEDIN, False)



from selenium.webdriver import Firefox
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import NoSuchElementException

import pytest

import time


@pytest.mark.needs_server
class TestMaxlifeFeature(object):
    """
    Checks if the maxlife feature is working
    """

    def setup_class(self):
        """
        Setup: Open a mozilla browser, login
        """
        self.browser = Firefox()
        self.browser.get('http://localhost:5000/')
        token = self.browser.find_element_by_name("token")
        password = "foo"
        # login
        token.send_keys(password)
        token.send_keys(Keys.ENTER)
        time.sleep(.1)
        try:
            self.browser.find_element_by_xpath("//input[@value='Logout']")
        except NoSuchElementException:
            raise ValueError("Can't login!!! Create a user 'foo' with the permissions"
                             "'read' and 'create' in your PERMISSIONS in the config")

    def teardown_class(self):
        """
        Tear down: Close the browser
        """
        self.browser.quit()

    @property
    def page_body_lowercase(self):
        return self.browser.find_element_by_tag_name("body").text.lower()

    def test_unit_input_exists(self):
        unit_input = self.browser.find_element_by_name("maxlife-unit")
        assert unit_input is not None
        value_input = self.browser.find_element_by_name("maxlife-value")
        assert value_input is not None

    def fill_form(self):
        """
        Fills test values to the form and submits it
        :return: tuple(filename, pasted_text)
        """
        filename = "test.txt"
        text_to_paste = "This is test"
        paste_input = self.browser.find_element_by_id("formupload")
        paste_input.send_keys(text_to_paste)
        filename_input = self.browser.find_element_by_id("filename")
        filename_input.send_keys(filename)
        contenttype_input = self.browser.find_element_by_id("contenttype")
        contenttype_input.send_keys("text/plain")
        contenttype_input.send_keys(Keys.ENTER)
        time.sleep(.2)  # give some time to render next view
        return filename, text_to_paste

    def delete_current_file(self):
        self.browser.find_element_by_id("del-btn").click()
        time.sleep(.2)
        self.browser.find_element_by_class_name("bootbox-accept").click()

    def test_paste_keep_forever(self):
        self.browser.find_element_by_xpath("//select[@name='maxlife-unit']/option[@value='forever']").click()
        value_input = self.browser.find_element_by_name("maxlife-value")
        value_input.clear()
        value_input.send_keys(1)
        self.fill_form()
        assert "max lifetime: forever" in self.page_body_lowercase
        self.delete_current_file()

    def test_paste_keep_minutes(self):
        self.browser.find_element_by_xpath("//select[@name='maxlife-unit']/option[@value='minutes']").click()
        value_input = self.browser.find_element_by_name("maxlife-value")
        value_input.clear()
        value_input.send_keys(1)
        self.fill_form()
        assert "max lifetime: forever" not in self.page_body_lowercase
        self.delete_current_file()

    def test_filename_gets_displayed(self):
        filename, _ = self.fill_form()
        assert filename.lower() in self.page_body_lowercase
        self.delete_current_file()

    def test_pasted_text_gets_displayed(self):
        _, pasted_text = self.fill_form()
        self.browser.find_element_by_id("inline-btn").click()
        assert pasted_text.lower() in self.page_body_lowercase
        self.browser.back()
        self.delete_current_file()

    @pytest.mark.slow
    def test_file_gets_deleted_after_expiry_time(self):
        self.browser.find_element_by_xpath("//select[@name='maxlife-unit']/option[@value='minutes']").click()
        value_input = self.browser.find_element_by_name("maxlife-value")
        value_input.clear()
        value_input.send_keys(1)
        self.fill_form()
        time.sleep(61)
        self.browser.find_element_by_id("inline-btn").click()
        assert "not found" in self.page_body_lowercase



from bepasty.storage.filesystem import Data


def test(tmpdir):
    p = tmpdir.join('test.data')

    d = Data(p.open('w+b'))
    assert d.size == 0

    d.write(b'a' * 1024, 0)
    assert d.size == 1024

    d.write(b'a' * 1024, 1024 * 3)
    assert d.size == 1024 * 4

    assert d.read(1024, 0) == b'a' * 1024
    assert d.read(1024, 1024) == b'\0' * 1024

    d.close()



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

this_dir = os.path.dirname(os.path.abspath(__file__))

readme_path = os.path.join(this_dir, 'README.rst')

with open(readme_path) as f:
    long_description = f.read()


setup(
    name='bepasty',
    use_scm_version={
        'write_to': 'src/bepasty/_version.py',
    },
    license='BSD 2-clause',
    author='The Bepasty Team (see AUTHORS file)',
    author_email='',
    maintainer='Thomas Waldmann',
    maintainer_email='tw@waldmann-edv.de',
    description='a binary pastebin / file upload service',
    long_description=long_description,
    url='https://github.com/bepasty/bepasty-server/',
    keywords="text image audio video binary pastebin upload download service wsgi flask",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'bepasty-server = bepasty.cli.server:main',
            'bepasty-object = bepasty.cli.object:main',
        ],
    },
    setup_requires=['setuptools_scm', ],
    install_requires=[
        'Flask',
        'Pygments',
        'xstatic',
        'XStatic-asciinema-player',
        'xstatic-bootbox>=5.4.0',
        'xstatic-bootstrap>=4.0.0.0,<5.0.0.0',
        'xstatic-font-awesome',
        'xstatic-jquery',
        'xstatic-jquery-ui',
        'xstatic-jquery-file-upload',
        'xstatic-pygments',
    ],
    extras_require={
        "magic": [
            'python-magic',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import os
import time
import hashlib

from flask import (
    Flask,
    Markup,
    current_app,
    g as flaskg,  # searching for 1 letter name "g" isn't nice, thus we use flaskg
    render_template,
    session,
)

from .apis import blueprint as blueprint_apis
from .storage import create_storage
from .utils.name import setup_werkzeug_routing
from .utils.permissions import (
    ADMIN,
    CREATE,
    DELETE,
    MODIFY,
    LIST,
    READ,
    get_permission_icons,
    get_permissions,
    logged_in,
    may,
)
from .views import blueprint

import mimetypes
mimetypes.add_type('application/x-asciinema-recording', '.cast')


class PrefixMiddleware(object):
    def __init__(self, app, prefix=''):
        self.app = app
        self.prefix = prefix

    def __call__(self, environ, start_response):
        if environ['PATH_INFO'].startswith(self.prefix):
            environ['PATH_INFO'] = environ['PATH_INFO'][len(self.prefix):]
            environ['SCRIPT_NAME'] = self.prefix
            return self.app(environ, start_response)
        else:
            start_response('404', [('Content-Type', 'text/plain')])
            return ['This URL does not belong to the bepasty app.'.encode()]


def setup_secret_key(app):
    """
    The secret key is used to sign cookies and cookies not signed with the
    current secret key are considered invalid.

    Here, we amend the configured secret key, so it depends on some
    other config values. Changing any of these values will change the
    computed secret key (and thus invalidate all previously made
    cookies).

    Currently supported secret-changing config values: PERMISSIONS
    """
    # if app.config['SECRET_KEY'] is empty, keep as NullSession
    if app.config['SECRET_KEY']:
        perms = sorted(k + v for k, v in app.config['PERMISSIONS'].items())
        perms = ''.join(perms).encode()
        app.config['SECRET_KEY'] += hashlib.sha256(perms).hexdigest()


def create_app():
    app = Flask(__name__)

    app.config.from_object('bepasty.config.Config')
    if os.environ.get('BEPASTY_CONFIG'):
        app.config.from_envvar('BEPASTY_CONFIG')

    setup_secret_key(app)

    prefix = app.config.get('APP_BASE_PATH')
    if prefix is not None:
        app.wsgi_app = PrefixMiddleware(app.wsgi_app, prefix=prefix)

    app.storage = create_storage(app)
    setup_werkzeug_routing(app)

    app.register_blueprint(blueprint)
    app.register_blueprint(blueprint_apis)

    @app.errorhandler(403)
    def url_forbidden(e):
        heading = 'Forbidden'
        body = Markup("""\
            <p>
                You are not allowed to access the requested URL.
            </p>
            <p>
                If you entered the URL manually please check your spelling and try again.
            </p>
            <p>
                Also check if you maybe forgot to log in or if your permissions are insufficient for this.
            </p>
""")
        return render_template('error.html', heading=heading, body=body), 403

    @app.errorhandler(404)
    def url_not_found(e):
        heading = 'Not found'
        body = Markup("""\
            <p>
                The requested URL was not found on the server.
            </p>
            <p>
                If you entered the URL manually please check your spelling and try again.
            </p>
""")
        return render_template('error.html', heading=heading, body=body), 404

    @app.before_request
    def before_request():
        """
        before the request is handled (by its view function), we compute some
        stuff here and make it easily available.
        """
        flaskg.logged_in = logged_in()
        flaskg.permissions = get_permissions()
        flaskg.icon_permissions = get_permission_icons()
        if flaskg.logged_in:
            session.permanent = current_app.config['PERMANENT_SESSION']

    def datetime_format(ts):
        """
        takes a unix timestamp and outputs a iso8601-like formatted string.
        times are always UTC, but we don't include the TZ here for brevity.
        it should be made clear (e.g. in the template) that the date/time is UTC.
        """
        if not ts:  # we use 0 to indicate undefined time
            return 'undefined'
        return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(ts))

    app.jinja_env.filters['datetime'] = datetime_format

    app.jinja_env.globals['flaskg'] = flaskg
    app.jinja_env.globals['may'] = may
    app.jinja_env.globals['ADMIN'] = ADMIN
    app.jinja_env.globals['LIST'] = LIST
    app.jinja_env.globals['CREATE'] = CREATE
    app.jinja_env.globals['MODIFY'] = MODIFY
    app.jinja_env.globals['READ'] = READ
    app.jinja_env.globals['DELETE'] = DELETE

    return app



import pytest
from werkzeug.exceptions import BadRequest

from bepasty.utils.http import ContentRange


def test_contentrange_parse():
    r = ContentRange.parse('bytes 0-0/2')
    assert r.begin == 0
    assert r.end == 0
    assert r.complete == 2
    assert r.size == 1
    assert not r.is_complete

    r = ContentRange.parse('bytes 0-1/2')
    assert r.begin == 0
    assert r.end == 1
    assert r.complete == 2
    assert r.size == 2
    assert r.is_complete

    with pytest.raises(BadRequest):
        ContentRange.parse('test 0-1/2')

    with pytest.raises(BadRequest):
        ContentRange.parse('bytes 1-0/2')

    with pytest.raises(BadRequest):
        ContentRange.parse('bytes 0-2/2')



from pygments.formatters.html import HtmlFormatter


class CustomHtmlFormatter(HtmlFormatter):
    """Custom HTML formatter. Adds an option to wrap lines into HTML <p> tags."""
    def __init__(self, **options):
        super(CustomHtmlFormatter, self).__init__(**options)
        self.lineparagraphs = options.get('lineparagraphs', '')

    def _wrap_lineparagraphs(self, inner):
        """Wrap lines into <p> tags

        :param inner: iterator of tuples of format (code, text)
        :return: iterator of tuples containing updated wrapped lines
        """
        s = self.lineparagraphs
        i = self.linenostart - 1
        for t, line in inner:
            if t:
                i += 1
                yield 1, '<p id="%s-%d">%s</p>' % (s, i, line)
            else:
                yield 0, line

    def format_unencoded(self, tokensource, outfile):
        """Format by wrapping pieces of text according to the user's options

        :param tokensource: iterator of tuples of format (code, text)
        :param outfile: output file handler
        """
        source = self._format_lines(tokensource)
        if self.hl_lines:
            source = self._highlight_lines(source)
        if not self.nowrap:
            if self.linenos == 2:
                source = self._wrap_inlinelinenos(source)
            if self.lineanchors:
                source = self._wrap_lineanchors(source)
            if self.linespans:
                source = self._wrap_linespans(source)
            if self.lineparagraphs:
                source = self._wrap_lineparagraphs(source)
            source = self.wrap(source, outfile)
            if self.linenos == 1:
                source = self._wrap_tablelinenos(source)
            if self.full:
                source = self._wrap_full(source, outfile)

        for t, piece in source:
            outfile.write(piece)



class Config(object):
    """
    This is the basic configuration class for bepasty.

    IMPORTANT:

    The config is only loaded at startup time of the app, so if you change it,
    you need to restart the wsgi app process(es) to make it load the updated
    config.
    """

    #: name of this site (put YOUR bepasty fqdn here)
    SITENAME = 'bepasty.example.org'

    #: base URL path of app (if not served on root URL, but e.g. on http://example.org/bepasty ).
    #: setting this to a non-None value will activate the PrefixMiddleware that splits PATH_INFO
    #: into SCRIPT_NAME (== APP_BASE_PATH) and the rest of PATH_INFO.
    APP_BASE_PATH = None  # '/bepasty'

    #: Whether files are automatically locked after upload.
    #:
    #: If you want to require admin approval and manual unlocking for each
    #: uploaded file, set this to True.
    UPLOAD_LOCKED = False

    #: The asciinema player theme (one of asciinema, tango, solarized-dark,
    #: solarized-light, monokai)
    ASCIINEMA_THEME = 'asciinema'

    #: The site admin can set some maximum allowed file size he wants to
    #: accept here. This is the maximum size an uploaded file may have.
    MAX_ALLOWED_FILE_SIZE = 5 * 1000 * 1000 * 1000

    #: The maximum http request body size.
    #: This is an information given to rest api clients so they can adjust
    #: their chunk size accordingly.
    #:
    #: This needs to be in sync with (or at least not beyond) the web server
    #: settings:
    #: apache:  LimitRequestBody 1048576 # apache default is 0 (unlimited)
    #: nginx:  client_max_body_size 1m; # nginx default (== 1048576)
    MAX_BODY_SIZE = 1 * 1024 * 1024 - 8192  # 8kiB safety margin, issue #155

    #: Setup maximum file sizes for specific content-types. If an item is
    #: beyond the limit set for its type, it will not be rendered, but just
    #: offered for download. Lookup within MAX_RENDER_SIZE is done by
    #: first-match and it is automatically sorted for longer content-type-
    #: prefixes first.
    #:
    #: Format of entries: content-type-prefix: max_size
    MAX_RENDER_SIZE = {
        # each list entry has 38 bytes, do not render > 1000 items
        'text/x-bepasty-list': 1000 * 38,
        # stuff rendered with syntax highlighting (expensive for server and
        # client) and also used for other text/* types as we use same code to
        # get a (non-highlighted) display with line numbers:
        'HIGHLIGHT_TYPES': 100 * 1000,
        # the in-browser pdf reader is sometimes rather slow and should
        # rather not be used for big PDFs:
        'application/pdf': 10 * 1000 * 1000,
        'application/x-pdf': 10 * 1000 * 1000,
        # images / audio / video can be rather big, we do not process them:
        'image/': 10 * 1000 * 1000,
        'audio/': 1 * 1000 * 1000 * 1000,
        'video/': 5 * 1000 * 1000 * 1000,
        # DEFAULT - matches everything not matched otherwise.
        # As we have catched everything we are able to render already,
        # this maybe should be a medium size, just for the case we forget
        # something:
        '': 1 * 1000 * 1000,
    }

    # Whether to use the python-magic module to guess a file's mime
    # type by looking into its content (if the mime type can not be
    # determined from the filename extension).
    # NOTE:
    # libmagic may have security issues, so maybe you should only use
    # it if you trust all users with upload permissions ('create').
    USE_PYTHON_MAGIC = False

    #: Define storage backend, choose from:
    #:
    #: - 'filesystem'
    #:
    STORAGE = 'filesystem'

    #: Filesystem storage path
    STORAGE_FILESYSTEM_DIRECTORY = '/tmp/'

    #: server secret key needed for safe session cookies.
    #: you must set a very long, very random, very secret string here,
    #: otherwise bepasty will not work (and crash when trying to log in)!
    SECRET_KEY = ''

    #: transmit cookie only over https (if you use http, set this to False)
    SESSION_COOKIE_SECURE = True
    #: use a permanent session (True, cookie will expire after the given
    #: time, see below) or not (False, cookie will get removed when browser
    #: is closed)
    PERMANENT_SESSION = False
    #: lifetime of the permanent session (in seconds)
    PERMANENT_SESSION_LIFETIME = 31 * 24 * 3600

    #: Bepasty does **not** use the usual user/password scheme, but **only**
    #: uses passwords (or passphrases - we'll call both "a secret" below) as
    #: log-in credentials - there are no separate user names.
    #:
    #: People who log-in using such a secret may get more permissions than
    #: those who do not log-in (who just get DEFAULT_PERMISSIONS).
    #:
    #: Depending on the secret used to log-in, they will get the permissions
    #: configured here, see below.
    #:
    #: You can use same secret / same permissions for all privileged users or
    #: set up different secrets / different permissions for each user.
    #:
    #: If you want to be able to revoke permissions for some user / some group
    #: of users, it might be a good idea to remember to whom you gave which
    #: secret (and also handle it in a rather fine-grained way).
    #:
    #: PERMISSIONS is a dict that maps secrets to permissions, use it like:
    #:
    #: ::
    #:
    #:     PERMISSIONS = {
    #:         'myadminsecret_1.21d-3!wdar34': 'admin,list,create,modify,read,delete',
    #:         'uploadersecret_rtghtrbrrrfsd': 'create,read',
    #:         'joe_doe_89359299887711335537': 'create,read,delete',
    #:     }
    PERMISSIONS = {
        # 'foo': 'admin,list,create,modify,read,delete',
    }

    #: not-logged-in users get these permissions -
    #: usually they are either no permissions ('') or read-only ('read').
    DEFAULT_PERMISSIONS = ''



import pytest

from bepasty.utils._compat import range_type
from bepasty.utils.name import ItemName, encode, make_id


def test_create():
    fake_storage = {}
    n = ItemName.create(fake_storage)
    assert n


def test_create_many():
    fake_storage = {}
    length = 1
    count = 400  # way more than we can do with this name length
    max_seen_length = 0
    for i in range_type(count):
        name = ItemName.create(fake_storage, length=length, max_length=length * 4, max_tries=10)
        # use the name in storage, so it is not available any more
        fake_storage[name] = None
        max_seen_length = max(max_seen_length, len(name))
    # it should automatically use longer names, if it runs out of unique names:
    assert max_seen_length > length
    # we should get all unique names we wanted, no duplicates:
    assert len(list(fake_storage)) == count


def test_make_id_type():
    assert isinstance(make_id(2), str)


def test_make_id_length():
    for length in range_type(10):
        assert len(make_id(length)) == length


def test_make_id_alphabet():
    # id must contain alphabet chars ONLY
    assert set(make_id(10, alphabet="abc")) <= set(['a', 'b', 'c'])


def test_make_id_unique():
    length, count = 6, 10000
    ids = set(make_id(length) for i in range_type(count))
    # if we did not encounter duplicates, set size must be <count>
    # of course, in extremely rare cases, this test might fail
    assert len(ids) == count


def test_encode_length():
    length = 42
    assert len(encode(12345, length)) == length


def test_encode_binary():
    assert encode(0, 0, "01") == []  # zero length
    assert encode(1, 0, "01") == []  # zero length
    assert encode(0, 1, "01") == ['0']  # length match
    assert encode(1, 1, "01") == ['1']  # length match
    assert encode(0, 2, "01") == ['0', '0']  # leading zeroes
    assert encode(1, 2, "01") == ['0', '1']  # leading zeroes
    assert encode(2, 2, "01") == ['1', '0']  # length match
    assert encode(3, 2, "01") == ['1', '1']  # length match
    assert encode(4, 2, "01") == ['0', '0']  # overflow truncated


def test_encode_special():
    # equivalent to binary, but we see the code is rather flexible
    assert encode(0, 2, ".+") == ['.', '.']
    assert encode(1, 2, ".+") == ['.', '+']
    assert encode(2, 2, ".+") == ['+', '.']
    assert encode(3, 2, ".+") == ['+', '+']


def test_encode_decimal():
    assert encode(123, 3, "0123456789") == ['1', '2', '3']  # length match
    assert encode(456, 4, "0123456789") == ['0', '4', '5', '6']  # leading zeroes
    assert encode(7890, 3, "0123456789") == ['8', '9', '0']  # overflow truncated


def test_encode_hex():
    assert encode(31, 2, "0123456789ABCDEF") == ['1', 'F']


def test_encode_errors():
    with pytest.raises(ValueError):  # ValueError: alphabet must be at least 2 elements long
        encode(1, 1, "0")

    with pytest.raises(ValueError):  # ValueError: length must be >= 0
        encode(1, -1, "0")

    with pytest.raises(ValueError):  # ValueError: negative values are not supported
        encode(-1, 1, "0")
